import pdb
import requests

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL

from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter

class TestRawResponseAdapter():

  def test_deterministic_request(self):
    res = requests.get(DETERMINISTIC_GET_REQUEST_URL, stream=True) 

    raw_response = PythonResponseAdapterFactory(res).raw_response()
    self.__test_equivalence(raw_response, res)

  def test_non_deterministic_request(self):
    res = requests.get(NON_DETERMINISTIC_GET_REQUEST_URL, stream=True) 

    raw_response = PythonResponseAdapterFactory(res).raw_response()
    self.__test_equivalence(raw_response, res)

  def __test_equivalence(self, raw_response: str, response: requests.Response):
    _response = RawHttpResponseAdapter(raw_response).to_response()
    
    assert _response.request == None
    assert _response.raw.data == response.raw.data
    assert _response.status_code == response.status_code

    for key, val in response.headers.items():
      assert val == _response.headers.get(key), key





import os
import pathlib
import pdb
import yaml

from .constants import APP_EGRESS_NETWORK, APP_EGRESS_NETWORK_NAME, APP_INGRESS_NETWORK, APP_INGRESS_NETWORK_NAME, DOCKER_COMPOSE_CUSTOM 

class Builder():

  def __init__(self, dir_path: str, compose_file_name: str):
    self.__compose_file_name = compose_file_name
    self.__dir_path = dir_path
    self.__networks = {}
    self.__services = {}
    self.__templates_dir = os.path.join(pathlib.Path(__file__).parent.parent.resolve(), 'templates')
    self.__volumes = {}

  @property
  def compose_file_name(self):
    return self.__compose_file_name

  @property
  def compose_file_path(self):
    return os.path.join(self.dir_path, self.compose_file_name)

  @property
  def custom_compose_file_path(self):
    return os.path.join(
      self.dir_path,
      DOCKER_COMPOSE_CUSTOM
    )

  @property
  def dir_path(self):
    return self.__dir_path

  @property
  def egress_network(self):
    return self.__networks.get(APP_EGRESS_NETWORK_NAME)

  @property
  def egress_network_name(self):
    return APP_EGRESS_NETWORK_NAME

  @property
  def networks(self):
    return self.__networks

  @property
  def services(self):
    return self.__services

  @property
  def templates_dir(self):
    return self.__templates_dir

  @property
  def volumes(self):
    return self.__volumes

  def build_extends(self, service_name: str, source_dir: str):
    return {
      'file': os.path.relpath(self.compose_file_path, source_dir),
      'service': service_name
    }

  # If the file already exists, load the specified resources
  def load(self):
    if not os.path.exists(self.compose_file_path):
      return

    with open(self.compose_file_path, 'r') as fp:
      contents = yaml.safe_load(fp) or {}

      networks = contents.get('networks')
      if isinstance(networks, dict):
        self.__networks = networks

      services = contents.get('services')
      if isinstance(services, dict):
        self.__services = services

      volumes = contents.get('volumes')
      if isinstance(volumes, dict):
        self.__volumes = volumes

  def with_network(self, network):
    self.__networks[network] = {
      'name': f"{APP_EGRESS_NETWORK}.{network}"
    }
    return self

  def with_egress_network(self, network = APP_EGRESS_NETWORK):
    self.__networks[APP_EGRESS_NETWORK_NAME] = {
      'external': True,
      'name': network,
    }
    return self

  def with_ingress_network(self, network = APP_INGRESS_NETWORK):
    self.__networks[APP_INGRESS_NETWORK_NAME] = {
      'external': True,
      'name': network,
    }
    return self

  def with_service(self, service_name: str, service: dict):
    self.__services[service_name] = service
    return self

  def with_volume(self, volume_name: str, volume: dict = {}):
    self.__volumes[volume_name] = volume
    return self

  def write(self, compose: dict, path = None):
    path = path or self.compose_file_path
    parent_dir = os.path.dirname(path)

    if not os.path.exists(parent_dir):
      os.makedirs(parent_dir)

    with open(path, 'w') as fp:
      yaml.dump(compose, fp, allow_unicode=True)
      fp.close()


import pdb

from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.models.scenario_model import ScenarioModel
from stoobly_agent.lib.api.keys.scenario_key import InvalidScenarioKey, ScenarioKey

from stoobly_agent.app.settings import Settings

def overwrite_scenario(scenario_key: str):
  if not scenario_key:
    return

  try:
    scenario_key = ScenarioKey(scenario_key)
  except InvalidScenarioKey:
    return

  settings = Settings.instance()
  scenario_model = ScenarioModel(settings) 
  scenario, status = scenario_model.show(scenario_key.id)

  if status != 200:
    return

  if not scenario.get('overwritable'):
    return

  status = scenario_model.update(scenario_key.id, **{ 'overwritable': False })[1]

  if status != 200:
    return 

  requests_model = RequestModel(settings)
  res, status = requests_model.destroy_all(scenario_id=scenario['id'])

  if status != 200:
    return

  return res


import pdb

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.lib.orm.request import Request

class PythonRequestAdapter():

  def __init__(self, request: Request):
    self.__request = request

  def adapt(self):
    raw_http_request_adapter = RawHttpRequestAdapter(self.__request.raw)
    return raw_http_request_adapter.to_request()


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.cli import config
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID

@pytest.fixture
def settings():
  return reset()

class TestMatch():

  def test_it_sets_rule(self, settings: Settings):
    runner = CliRunner()

    method = 'POST'
    _mode = mode.MOCK
    pattern = '.*?'

    match_result = runner.invoke(config, [
        'match', 'set', 
        '--method', method, '--mode', _mode, '--pattern', pattern, '--component', request_component.BODY_PARAM
      ]
    )
    assert match_result.exit_code == 0

    settings.load()

    match_rules =  settings.proxy.match.match_rules(str(LOCAL_PROJECT_ID))
    assert len(match_rules) == 1

    match_rule = match_rules[0]
    assert match_rule.pattern == pattern
    assert match_rule.methods == [method]
    assert match_rule.components == [request_component.BODY_PARAM]
    assert match_rule.modes == [_mode]


import click
import pdb
import sys

from stoobly_agent import VERSION
from stoobly_agent.app.proxy.replay.body_parser_service import decode_response, is_traversable
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.lib.api.response_param_names_resource import ResponseParamNamesResource
from stoobly_agent.lib.orm.migrate_service import migrate as database_migrate, rollback as database_rollback
from stoobly_agent.lib.orm.replayed_response import ReplayedResponse
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.orm.trace import Trace
from stoobly_agent.lib.orm.trace_alias import TraceAlias
from stoobly_agent.lib.orm.trace_request import TraceRequest
from stoobly_agent.lib.utils import jmespath

from stoobly_agent.app.models.adapters.orm.joined_request_string_adapter import JoinedRequestStringAdapter
from stoobly_agent.app.models.helpers.create_request_params_service import build_params

@click.group(
  epilog="Run 'stoobly-agent dev-tools COMMAND --help' for more information on a command.",
  help="Access developer tools"
)
@click.pass_context
def dev_tools(ctx):
    pass

@dev_tools.command()
def debug(**kwargs):
  pdb.set_trace()

@dev_tools.command()
def migrate():
  database_migrate(VERSION)

@dev_tools.command()
def rollback():
  database_rollback()

@dev_tools.command()
@click.option('--content')
@click.option('--content-type', default='application/json')
@click.argument('query')
def query(**kwargs):
  decoded_response = decode_response(kwargs['content'], kwargs['content_type'])
  
  if not is_traversable(decoded_response):
    print(f"{decoded_response} is not traversible", file=sys.stderr)
    sys.exit(1)

  print(jmespath.search(kwargs['query'], decoded_response))


from typing import List, TypedDict

DEFAULT_FORMAT = 'default'
JSON_FORMAT = 'json'

class Test(TypedDict):
  log: str
  passed: bool

class TestResponse(TypedDict):
  content: str
  latency: int
  status_code: int

class TestOutput(TypedDict):
  error: str
  expected_response: TestResponse
  passed: bool
  response: TestResponse
  tests: List[Test]

class ReplayOutput(TypedDict):
  content: str
  headers: dict
  method: str
  url: str


AGENT_ACTIVE_MODE = 'STOOBLY_AGENT_ACTIVE_MODE'
AGENT_CONFIG_PATH = 'STOOBLY_AGENT_CONFIG_PATH'
AGENT_ENABLED = 'STOOBLY_AGENT_ENABLED'
AGENT_SELF_INTERCEPT_ENABLED = 'STOOBLY_AGENT_SELF_INTERCEPT_ENABLED'
AGENT_INCLUDE_PATTERNS = 'STOOBLY_AGENT_INCLUDE_PATTERNS'
AGENT_IS_HEADLESS = 'STOOBLY_AGENT_IS_HEADLESS'
AGENT_EXCLUDE_PATTERNS = 'STOOBLY_AGENT_EXCLUDE_PATTERNS'
AGENT_LIFECYCLE_HOOKS_PATH = 'STOOBLY_AGENT_LIFECYCLE_HOOKS_PATH'
AGENT_POLICY = 'STOOBLY_AGENT_POLICY'
AGENT_PROJECT_KEY = 'STOOBLY_AGENT_PROJECT_KEY'
AGENT_PROXY_URL = 'STOOBLY_AGENT_PROXY_URL'
AGENT_PUBLIC_DIRECTORY_PATH = 'STOOBLY_AGENT_PUBLIC_DIRECTORY_PATH'
AGENT_REMOTE_ENABLED = 'STOOBLY_AGENT_REMOTE_ENABLED'
AGENT_REMOTE_PROJECT_KEY = 'STOOBLY_AGENT_REMOTE_PROJECT_KEY'
AGENT_REPLAY_HOST = 'STOOBLY_AGENT_REPLAY_HOST'
AGENT_REPLAY_SCHEME = 'STOOBLY_AGENT_REPLAY_SCHEME'
AGENT_RESPONSE_FIXTURES_PATH = 'STOOBLY_AGENT_RESPONSE_FIXTURES_PATH'
AGENT_SERVICE_URL = 'STOOBLY_AGENT_SERVICE_URL'
AGENT_SCENARIO_KEY = 'STOOBLY_AGENT_SCENARIO_KEY'
AGENT_SIMULATE_LATENCY = 'STOOBLY_AGENT_SIMULATE_LATENCY'
AGENT_URL = 'STOOBLY_AGENT_URL'
API_URL = 'STOOBLY_API_URL'
API_KEY = 'STOOBLY_API_KEY'
ENV = 'STOOBLY_AGENT_ENV'
FEATURE_REMOTE = 'STOOBLY_FETURE_REMOTE'
HTTP_PROXY = 'HTTP_PROXY'
HTTPS_PROXY = 'HTTPS_PROXY'
LOG_LEVEL = 'STOOBLY_AGENT_LOG_LEVEL'
NO_PROXY = 'NO_PROXY'


import pdb

from .python_adapter import PythonRequestAdapter

class RawJoinedRequestAdapterFactory():

  def __init__(self, joined_request: bytes):
    self.__request = joined_request

  def python_request(self):
    return PythonRequestAdapter(self.__request).adapt()


import os

from stoobly_agent.config.constants import env_vars

from .types.ui_settings import UISettings as IUISettings

class UISettings:

  def __init__(self, ui_settings: IUISettings):
    self.__ui_settings = ui_settings

    self.__active = self.active_before_change
    self.__url = self.url_before_change

  @property
  def active_before_change(self) -> bool:
    return self.__ui_settings.get('active')

  @property
  def active(self) -> bool:
    if self.__active != self.active_before_change:
      return self.__active

    if os.environ.get(env_vars.AGENT_IS_HEADLESS):
        return True

    return self.__active or False

  @active.setter
  def active(self, v):
    self.__active = v

  @property
  def url_before_change(self) -> str:
    return self.__ui_settings.get('url')

  @property
  def url(self) -> str:
    if self.__url != self.url_before_change:
        return self.__url

    if os.environ.get(env_vars.AGENT_URL):
        return os.environ[env_vars.AGENT_URL]

    return self.__url or ''

  @url.setter
  def url(self, v):
    self.__url = v

  def to_dict(self):
    return {
      'active': self.__active,
      'url': self.__url,
    }


import requests
import urllib

from stoobly_agent.app.models.types import ScenarioCreateParams

from ..logger import Logger
from .interfaces.pagination_query_params import PaginationQueryParams
from .stoobly_api import StooblyApi

LOG_ID = 'ScenariosResource'

class ScenariosResource(StooblyApi):
  SCENARIOS_ENDPOINT = 'scenarios'

  def create(self, **params: ScenarioCreateParams):
    url = f"{self.service_url}/{self.SCENARIOS_ENDPOINT}"
    return self.post(url, headers=self.default_headers, data=params)

  def index(self, **query_params: PaginationQueryParams) -> requests.Response:
    url = f"{self.service_url}/{self.SCENARIOS_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def show(self, scenario_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.SCENARIOS_ENDPOINT}/{scenario_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def update(self, scenario_id: int, **params):
    url = f"{self.service_url}/{self.SCENARIOS_ENDPOINT}/{scenario_id}"

    Logger.instance(LOG_ID).debug(f"{url}")

    return self.put(url, headers=self.default_headers, json=params)


from base64 import b64encode
import pdb

from httptools import parse_url
from typing import List
from urllib.parse import parse_qs

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.app.models.types import RequestShowParams
from stoobly_agent.lib.api.interfaces import QueryParam, RequestShowResponse
from stoobly_agent.lib.utils.decode import decode

from ..request import Request as ORMRequest
from ..response import Response
from .orm_to_stoobly_response_transformer import ORMToStooblyResponseTransformer

FILTER_LIST = [
  'headers_hash',
  'body_text_hash',
  'query_params_hash',
  'body_params_hash',
  'control',
  'raw',
] 

class ORMToStooblyRequestTransformer():
  __options: RequestShowParams
  __request: ORMRequest = None

  # TODO: filter by options
  def __init__(self, request: ORMRequest, options: RequestShowParams):
    self.__options = options
    self.__request = request

  def transform(self) -> RequestShowResponse:
    stoobly_request: RequestShowResponse = self.__request.to_dict()

    self.__decorate_with_request(stoobly_request, self.__request)
    self.__decorate_with_response(stoobly_request, self.__request.response)
    self.__decorate_with_components(stoobly_request, self.__request)
    self.__decorate_with_scenario(stoobly_request, self.__request)
    
    return self.__filter_properties(stoobly_request)

  def __filter_properties(self, stoobly_request: RequestShowResponse):
    allowed_keys = list(RequestShowResponse.__annotations__.keys()) + [
      'pushed_at', 'body_params_hash', 'body_text_hash', 'key', 'query', 'query_params_hash'
    ]

    filter_keys = lambda request: dict((key, value) for key, value in request.items() if key in allowed_keys)
    return filter_keys(stoobly_request)

  def __decorate_with_scenario(self, stoobly_request: RequestShowResponse, request: ORMRequest):
    if not request.scenario_id:
      return

    scenario = request.scenario

    if scenario:
      stoobly_request['scenario'] = scenario.name

  def __decorate_with_components(self, stoobly_request: RequestShowResponse, request: ORMRequest):
    components = []
    if len(request.query_params_hash) != 0:
      components.append('query_params')

    if len(request.body_params_hash) != 0:
      components.append('body_params')

    if len(request.body_text_hash) != 0:
      components.append('body')

    if len(request.headers_hash) != 0:
      components.append('headers')

    if len(request.response_hash) != 0:
      components.append('response')

    if len(request.response_headers_hash) != 0:
      components.append('response_headers')

    stoobly_request['components'] = components

  def __decorate_with_request(self, stoobly_request: RequestShowResponse, request: ORMRequest):
    python_request = RawHttpRequestAdapter(request.raw).to_request()

    stoobly_request['method'] = python_request.method
    stoobly_request['url'] = python_request.url
    stoobly_request['is_deleted'] = bool(stoobly_request.get('is_deleted'))

    parsed_url = parse_url(python_request.url.encode())

    stoobly_request['scheme'] = parsed_url.schema.decode()

    if parsed_url.query == None:
      if 'query_params' in self.__options:
        stoobly_request['query_params'] = [] 
    else:
      stoobly_request['query'] = decode(parsed_url.query)

      if 'query_params' in self.__options:
        stoobly_request['query_params'] = self.__transform_query_params(stoobly_request['query'])

    if 'headers' in self.__options:
      stoobly_request['headers'] = self.__transform_headers(python_request.headers)

    if 'body' in self.__options:
      stoobly_request['body'] = b64encode(python_request.data)

  def __transform_headers(self, headers: dict):
    headers_list = []
    for key, val in headers.items():
      headers_list.append({
        'name': key,
        'value': val,
      })
    return headers_list

  def __transform_query_params(self, query_params):
    query_params_list: List[QueryParam] = []

    for key, val in parse_qs(query_params).items():
      query_params_list.append({
        'name': key,
        'value': val,
      })
    return query_params_list
  
  def __decorate_with_response(self, stoobly_request: RequestShowResponse, orm_response: Response):
    if not orm_response:
      return

    transformer = ORMToStooblyResponseTransformer(orm_response)
    stoobly_request['status'] = self.__request.status

    if 'response' in self.__options:
      stoobly_response = transformer.transform()
      stoobly_request['response'] = stoobly_response



import pdb

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.query_param_model import QueryParamModel
from stoobly_agent.app.settings import Settings

class QueryParamsController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # POST /requests/:requestId/query_params
    def create(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1,
            'queryParamId': 3,
        })

        if not context.params.get('name'):
            return context.bad_request('Missing name')

        if not context.params.get('value'):
            return context.bad_request('Missing value')

        query_param_model = self.__query_param_model(context)
        query_param, status = query_param_model.create(context.params.get('requestId'), **{
            'name': context.params['name'],
            'value': context.params['value'],
        })

        if context.filter_response(query_param, status):
            return

        context.render(
            json = query_param,
            status = 200
        )

    # GET /requests/:requestId/query_params
    def index(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1
        })

        query_param_model = self.__query_param_model(context)
        query_params, status = query_param_model.index(context.params.get('requestId'), **context.params)

        if context.filter_response(query_params, status):
            return

        context.render(
            json = query_params,
            status = 200
        )

    # PUT /requests/:requestId/query_params/:queryParamId
    def update(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1,
            'queryParamId': 3,
        })

        if not context.params.get('name'):
            return context.bad_request('Missing name')

        if not context.params.get('value'):
            return context.bad_request('Missing value')

        query_param_model = self.__query_param_model(context)
        query_param, status = query_param_model.update(context.params.get('requestId'), context.params.get('queryParamId'), **{
            'name': context.params['name'],
            'value': context.params['value'],
        })

        if context.filter_response(query_param, status):
            return

        context.render(
            json = query_param,
            status = 200
        )

    # DELETE /requests/:requestId/query_params/:queryParamId
    def destroy(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'requestId': 1,
            'queryParamId': 3,
        })

        query_param_model = self.__query_param_model(context)
        query_param, status = query_param_model.destroy(context.params.get('requestId'), context.params.get('queryParamId'))

        if context.filter_response(query_param, status):
            return
        
        context.render(
            plain = query_param,
            status = status
        )

    def __query_param_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        query_param_model = QueryParamModel(Settings.instance(), access_token=access_token)
        return query_param_model


import pdb

from typing import Union

from stoobly_agent.lib.api.interfaces.endpoints import RequestComponentName
from stoobly_agent.lib.utils.python_to_ruby_type import convert

class SchemaBuilder:
  ###
  #
  # @param endpoint_id [Integer]
  # @param param_column_name [String] e.g. query_param_name, body_param_name, response_param_name
  #
  def __init__(self, endpoint_id, param_column_name: str):
    self.endpoint_id = endpoint_id
    self.param_column_name: str = param_column_name

  def build(self, params):
    params_list = []
    for literal_param in params:
      param: RequestComponentName = {
        'endpoint_id': self.endpoint_id,
        'name': literal_param['name'],
        'query': literal_param['query'],
        'is_required': literal_param['required'],
        'inferred_type': convert(self.__infer_type(literal_param['value'])),
        'is_deterministic': True,
        'id': literal_param['id'],
        f"{self.param_column_name}_id": literal_param['parent_id']
      }
      params_list.append(param)
    return params_list

  def __infer_type(self, val) -> str:
    return str(val.__class__)



import pdb
import requests

from stoobly_agent.app.settings import Settings

from .factories.resource.body import BodyResourceFactory
from .model import Model

class BodyModel(Model):

  def __init__(self, settings: Settings, **options):
    super().__init__(settings, **options)

  def as_local(self):
    self.adapter = BodyResourceFactory(self.settings.remote).local_db()
    self.is_local = False

  def as_remote(self):
    self.is_local = False

  def update(self, request_id: int, text: str):
    try:
      return self.adapter.update(request_id, text)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def mock(self, request_id: str):
    try:
      return self.adapter.mock(request_id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)


import json
import pdb
import re

from mitmproxy.http import Headers, Request as MitmproxyRequest
from mitmproxy.coretypes import multidict
from typing import Callable, List
from urllib.parse import urlparse

from stoobly_agent.app.settings.constants import request_component
from stoobly_agent.app.settings.rewrite_rule import ParameterRule, RewriteRule, UrlRule 
from stoobly_agent.config.constants import custom_headers
from stoobly_agent.lib.logger import Logger, bcolors
from stoobly_agent.lib.utils import jmespath
from stoobly_agent.lib.utils.decode import decode

from .request_body_facade import MitmproxyRequestBodyFacade
from .request import Request

LOG_ID = 'Request'

class MitmproxyRequestFacade(Request):

    ###
    #
    # @param request [ActionDispatch::Request]
    #
    # @return [Hash]
    #
    def __init__(self, request: MitmproxyRequest):
        self.__url_rules: List[UrlRule] = []
        self.__parameter_rules: List[ParameterRule] = []

        self.request = request
        self.uri = urlparse(self.request.url)
        self.__body = MitmproxyRequestBodyFacade(request)

    @property
    def http_version(self):
        return self.request.http_version

    @property
    def url(self):
        uri = self.uri._replace(path=self.uri.path.rstrip('/'))
        return uri.geturl()

    @property
    def password(self):
        return self.uri.password

    @property
    def username(self):
        return self.uri.username

    @property
    def path(self):
        return self.uri.path

    @property
    def base_url(self):
        return f"{self.request.scheme}://{self.request.host}:{self.request.port}"

    @property
    def method(self):
        return self.request.method

    @property
    def headers(self):
        return self.__filter_custom_headers(self.request.headers)

    @property
    def body(self):
        content = self.request.raw_content or b''

        return decode(content)

    @property
    def parsed_body(self):
        return self.__body.get(self.content_type)

    @property
    def query(self):
        return self.request.query

    @property
    def query_string(self):
        from urllib.parse import quote

        params = []
        for k in self.query.keys():
            for v in self.query.get_all(k):
                params.append(f"{quote(k)}={quote(v)}")
        
        return '&'.join(params)

    @property
    def content_type(self):
        return self.headers.get('content-type')

    @property
    def host(self):
        return self.request.host

    @property
    def port(self):
        return self.request.port

    @property
    def url_rules(self) -> List[ParameterRule]:
        return self.__url_rules
    
    @property
    def parameter_rules(self) -> List[ParameterRule]:
        return self.__parameter_rules

    @property
    def scheme(self):
        return self.request.scheme

    def with_parameter_rules(self, rules: List[RewriteRule]):
        if type(rules) == list:
            self.__parameter_rules = self.select_parameter_rules(rules)
        return self 

    def with_url_rules(self, rules: List[RewriteRule]):
        if type(rules) == list:
            self.__url_rules = self.select_url_rules(rules)
        return self

    def rewrite(self):
        rewrites = self.__parameter_rules

        if len(rewrites) != 0:
            self.__rewrite_headers(rewrites)
            self.__rewrite_query(rewrites)
            self.__rewrite_content(rewrites)

        rewrites = self.__url_rules

        if len(rewrites):
            url_before = self.url
            self.__rewrite_url(rewrites)
            Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN} Rewriting URL{bcolors.ENDC} {url_before} => {self.url}")

    # Find all the rules that match request url and method
    def select_rewrite_rules(self, rules: List[RewriteRule]) -> List[RewriteRule]:
        return list(filter(self.__is_rewrite_rule_selected, rules or []))

    def select_parameter_rules(self, rules: List[RewriteRule]) -> List[ParameterRule]:
        _rules = self.select_rewrite_rules(rules)
        
        if len(_rules) == 0:
            return []

        parameter_rules = list(map(lambda rule: rule.parameter_rules, _rules))

        return [item for sublist in parameter_rules for item in sublist] # flatten list

    def select_url_rules(self, rules: List[RewriteRule]) -> List[UrlRule]:
        _rules = self.select_rewrite_rules(rules)

        if len(_rules) == 0:
            return []

        url_rules = list(map(lambda rule: rule.url_rules, _rules))

        return [item for sublist in url_rules for item in sublist] # flatten list

    def __is_rewrite_rule_selected(self, rewrite_rule: RewriteRule):
        pattern = rewrite_rule.pattern

        try:
            url_matches = re.match(pattern, self.url)
        except re.error as e:
            Logger.instance().error(f"RegExp error '{e}' for {pattern}")
            return False

        method = self.method.upper()
        method_matches = method in rewrite_rule.methods
        return url_matches and method_matches
    
    def __apply_rewrites(self, params: dict, rewrites: List[ParameterRule], handler: Callable):
        if len(rewrites) == 0:
            return

        for rewrite in rewrites:
            jmespath.search(rewrite.name, params, {
                'replacements': [handler(rewrite) if handler else rewrite.value],
            })

    def __rewrite_handler(self, rewrite: ParameterRule) -> str:
        Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Rewriting{bcolors.ENDC} {rewrite.type.lower()} {rewrite.name} => {rewrite.value}")
        return rewrite.value

    def __rewrite_url(self, rewrites: List[UrlRule]):
        for rewrite in rewrites:
            if rewrite.hostname:
                self.request.host = rewrite.hostname
            
            if rewrite.port:
                self.request.port = int(rewrite.port)

            if rewrite.scheme:
                self.request.scheme = rewrite.scheme.lower()

            if rewrite.path:
                path = rewrite.path
                if path[0] != '/':
                    path = '/' + path

                if self.uri.query:
                    self.request.path = f"{path}?{self.uri.query}"
                else:
                    self.request.path = path

        self.uri = urlparse(self.request.url)

    def __rewrite_headers(self, rewrites: List[ParameterRule]):
        self.__apply_headers(rewrites, self.__rewrite_handler)

    def __rewrite_query(self, rewrites: List[ParameterRule]):
        self.__apply_queries(rewrites, self.__rewrite_handler)

    def __rewrite_content(self, rewrites: List[ParameterRule]):
        self.__apply_content(rewrites, self.__rewrite_handler)

    def __apply_headers(self, rewrites: List[ParameterRule], handler: Callable):
        rewrites = list(filter(lambda rewrite: rewrite.type == request_component.HEADER, rewrites))
        self.__apply_rewrites(self.request.headers, rewrites, handler)

    def __apply_queries(self, rewrites: List[ParameterRule], handler: Callable):
        rewrites = list(filter(lambda rewrite: rewrite.type == request_component.QUERY_PARAM, rewrites))
        self.__apply_rewrites(self.request.query, rewrites, handler)

    def __apply_content(self, rewrites: List[ParameterRule], handler: Callable):
        rewrites = list(filter(lambda rewrite: rewrite.type == request_component.BODY_PARAM, rewrites))
        if len(rewrites) == 0:
            return

        content_type = self.content_type
        parsed_content = self.__body.get(content_type)

        if not self.__is_iterable(parsed_content):
            content_type = 'application/json'
            self.request.headers['content-type'] = content_type
            parsed_content = {}

        self.__apply_rewrites(parsed_content, rewrites, handler)
        self.__body.set(parsed_content, content_type)

    def __filter_custom_headers(self, request_headers: Headers):
        '''
        Remove custom headers
        '''
        _request_headers = Headers(**request_headers)

        headers = custom_headers.__dict__
        for key in headers:
            if key[0:2] == '__' and key[-2:] == '__':
                continue

            name = headers[key]

            if name not in request_headers:
                continue

            _request_headers.pop(name)

        return _request_headers

    def __is_iterable(self, v):
        return isinstance(v, dict) or isinstance(v, multidict.MultiDictView) or isinstance(v, list)


import json
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, request
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

@pytest.fixture(autouse=True, scope='module')
def settings():
  return reset()

class TestResponseGet():

  @pytest.fixture(autouse=True, scope='class')
  def request_create_cli_result(self, runner: CliRunner):
    record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
    assert record_result.exit_code == 0
    return record_result

  def test_it_returns_correct_response(self, runner: CliRunner):
    _request = Request.last()

    request_result = runner.invoke(request, ['response', 'get', _request.key()])
    assert request_result.exit_code == 0

    output = request_result.stdout
    assert type(output) is str
    assert type(output) is not bytes

    # Note: the invoke result's stdout can actually return a string of
    # a byte literal so parse the json for extra validation
    json_response = json.loads(output)
    assert json_response['message'] is not None
    assert json_response['status'] == 'success'




from .resource_key import ResourceKey
from .organization_key import LOCAL_ORGANIZATION_ID

LOCAL_PROJECT_ID = 0

class InvalidProjectKey(Exception):
  pass

class ProjectKey(ResourceKey):

  def __init__(self, key: str):
    super().__init__(key)

    if not self.id:
      raise InvalidProjectKey('Missing id')

    if not self.organization_id:
      raise InvalidProjectKey('Missing organization_id')

  @property
  @classmethod
  def local_key(cls) -> str:
    return cls.encode({
      'i': LOCAL_PROJECT_ID,
      'o': LOCAL_ORGANIZATION_ID,
    })

  @property
  def is_local(self) -> bool:
    return int(self.id) == LOCAL_PROJECT_ID

  @property
  def id(self) -> str:
    return self.get('i')
  
  @property
  def organization_id(self) -> str:
    return self.get('o')

  @staticmethod
  def check_is_local(id):
    return int(id) == LOCAL_PROJECT_ID

  @staticmethod
  def encode(id, organization_id = -1):
    return super(ProjectKey, ProjectKey).encode({
      'i': id,
      'o': organization_id,
    })




class ManagedServicesDockerCompose():
  def __init__(self, target_workflow_name):
    self.init_container_name = f"{target_workflow_name}-build.init-1"
    self.configure_container_name = f"{target_workflow_name}-build.configure-1"
    self.gateway_container_name = f"{target_workflow_name}-gateway.service-1"
    self.mock_ui_container_name = f"{target_workflow_name}-stoobly_ui.service-1"
    self.entrypoint_init_container_name = f"{target_workflow_name}-entrypoint.init-1"
    self.entrypoint_configure_container_name = f"{target_workflow_name}-entrypoint.configure-1"



ALIAS_RESOLVE_STRATEGY = 'X-Stoobly-Alias-Resolve-Strategy'
CONTENT_TYPE = 'X-Stoobly-Content-Type'
CONTENT_TYPE_TEST_RESULTS = 'test/results'
REMOTE_PROJECT_KEY = 'X-Stoobly-Endpoints-Project-Id'
MOCK_POLICY = 'X-Stoobly-Mock-Policy'
MOCK_REQUEST_ID = 'X-Stoobly-Request-Id'
MOCK_REQUEST_ENDPOINT_ID = 'X-Stoobly-Request-Endpoint-Id'
MOCK_REQUEST_KEY = 'X-Stoobly-Request-Key'
DO_PROXY = 'X-Stoobly-Do-Proxy'
LIFECYCLE_HOOKS_PATH = 'X-Stoobly-Lifecycle-Hooks-Path'
PROJECT_KEY = 'X-Stoobly-Project-Key'
PROXY_MODE = 'X-Stoobly-Proxy-Mode'
PUBLIC_DIRECTORY_PATH = 'X-Stoobly-Public-Directory-Path'
RECORD_POLICY = 'X-Stoobly-Record-Policy'
REPORT_KEY = 'X-Stoobly-Report-Key'
REQUEST_ORIGIN = 'X-Stoobly-Request-Origin'
RESPONSE_FIXTURES_PATH = 'X-Stoobly-Response-Fixtures-Path'
RESPONSE_ID = 'X-Stoobly-Response-Id'
RESPONSE_LATENCY = 'X-Stoobly-Request-Response-Latency'
RESPONSE_PROXY_MODE = 'X-Stoobly-Response-Proxy-Mode'
SCENARIO_KEY = 'X-Stoobly-Scenario-Key'
SESSION_ID = 'X-Stoobly-Session-Id'
SERVICE_URL = 'X-Stoobly-Service-Url'
TEST_FILTER = 'X-Stoobly-Test-Filter'
TEST_ID = 'X-Stoobly-Test-Id'
TEST_SAVE_RESULTS = 'X-Stoobly-Test-Save-Results'
TEST_SKIP = 'X-Stoobly-Test-Skip'
TEST_STRATEGY = 'X-Stoobly-Test-Strategy'
TRACE_ID = 'X-Stoobly-Trace-Id'
TRACE_REQUEST_ID = 'X-Stoobly-Trace-Request-Id'


import pdb
import requests

from typing import Union

from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.cli.helpers.test_facade import TestFacade
from stoobly_agent.app.proxy.replay.body_parser_service import decode_response
from stoobly_agent.app.proxy.test.helpers.test_results_builder import TestResultsBuilder
from stoobly_agent.app.proxy.test.helpers.diff_service import diff
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import custom_headers
from stoobly_agent.lib.api.interfaces.tests import TestShowResponse
from stoobly_agent.lib.api.keys.project_key import InvalidProjectKey, ProjectKey
from stoobly_agent.lib.utils.decode import decode

class TestReplayContext(ReplayContext):

  def __init__(self, replay_context: ReplayContext):
    super().__init__(replay_context.request)

    self.with_response(replay_context.response)
    self.end_time = replay_context.end_time
    self.start_time = replay_context.start_time

    self.__project_id = -1
    self.__test_results_builder = TestResultsBuilder()

    if self.has_test_results:
      self.__test_results_builder.deserialize(replay_context.response.content)

  @property
  def has_test_results(self):
    headers = self.response.headers
    return headers.get(custom_headers.CONTENT_TYPE) == custom_headers.CONTENT_TYPE_TEST_RESULTS

  @property
  def project_id(self):
    return self.__project_id

  @project_id.setter
  def project_id(self, v):
    self.__project_id = v

  @property
  def response_content(self):
    if self.has_test_results:
      return self.__test_results_builder.received_response
    else:
      content = self.response.content
      content_type = self.response.headers.get('content-type')

      return decode_response(content, content_type)

  @property
  def response_status_code(self):
    if self.has_test_results:
      return self.__test_results_builder.status
    else:
      return self.response.status_code

  @property
  def test_results_builder(self):
    return self.__test_results_builder

  def diff_response(self, response: TestShowResponse, test_facade: TestFacade):
    if response['passed'] or response['skipped']:
      return (self.response_content, self.response_status_code, self.latency)

    expected_response_content, expected_status_code = self.expected_response_content(test_facade)

    if not self.has_test_results:
      return (
        diff(expected_response_content, self.response_content),
        diff(expected_status_code, self.response_status_code),
        diff(response['expected_latency'], self.latency)
      )
    else:
      builder = self.test_results_builder
      return (
        diff(expected_response_content, builder.received_response),
        diff(expected_status_code, self.response_status_code),
        diff(response['expected_latency'], self.latency)
      )

  def expected_response_content(self, test_facade: TestFacade):
    if not self.has_test_results: 
      response = self.__get_test_expected_response_with_context(test_facade)

      if isinstance(response, requests.Response):
        content = decode(response.content)
        content_type = response.headers.get('content-type')

        return decode_response(content, content_type), response.status_code
      else:
        return '', 0
    else:
      builder = self.test_results_builder
      return builder.expected_response, builder.expected_status_code

  def test_show_response(self, test_facade = None) -> TestShowResponse:
    if not self.has_test_results: 
      return self.__get_test_response_with_context(test_facade)
    else:
      builder = self.test_results_builder

      # The following is inferred from settings (assumes nothing has changed)
      settings = Settings.instance()
      project_id = -1

      try:
        project_key = ProjectKey(settings.proxy.intercept.project_key) 
        project_id = project_key.id
      except InvalidProjectKey:
        pass

      return {
        'expected_latency': builder.expected_latency,
        'id': None,
        'log': builder.log,
        'passed': builder.passed,
        'project_id': project_id,
        'skipped': builder.skipped,
        'strategy': builder.strategy,
      }
    
  def __get_test_response_with_context(self, test_facade: TestFacade) -> Union[TestShowResponse, str]:
    try:
      res = test_facade.show_with_context(super(), self.project_id)
      return res
    except AssertionError as e:
      return f"API Error: {e}"

  def __get_test_expected_response_with_context(self, test_facade: TestFacade):
    try:
      res = test_facade.expected_response_with_context(super(), self.project_id)
      return res
    except AssertionError as e:
      return f"API Error: {e}"


import os
import pdb

from urllib.parse import urlparse

from ...constants import SERVICE_HOSTNAME, SERVICE_PORT, SERVICE_PROXY_MODE, STOOBLY_CERTS_DIR
from .builder import WorkflowBuilder

class ReverseProxyDecorator():

  def __init__(self, workflow_builder: WorkflowBuilder):
    self.__workflow_builder = workflow_builder

  @property
  def workflow_builder(self):
    return self.__workflow_builder

  @property
  def service_builder(self):
    return self.workflow_builder.service_builder

  def decorate(self):
    config = self.service_builder.config

    command = [
      '--headless',
      '--lifecycle-hooks-path', 'lifecycle_hooks.py',
      '--proxy-mode', SERVICE_PROXY_MODE,
      '--proxy-port', f"{SERVICE_PORT}",
      '--ssl-insecure'
    ]

    if config.scheme == 'https':
      command.append('--certs')
      command.append(os.path.join(STOOBLY_CERTS_DIR, f"{SERVICE_HOSTNAME}-joined.pem"))

    services = self.workflow_builder.services
    proxy_name = self.workflow_builder.proxy
    proxy_service = services.get(proxy_name) or {}

    additional_properties = { 'command': command }

    service = { 
      **proxy_service,
      **additional_properties,
    }

    services[proxy_name] = service 




import requests
import urllib
import pdb

from ..logger import Logger
from .interfaces import EndpointsIndexQueryParams
from .stoobly_api import StooblyApi

LOG_ID = 'EndpointsResource'

class EndpointsResource(StooblyApi):
  ENDPOINTS_ENDPOINT = 'endpoints'

  def create(self, **params):
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}"
    return self.post(url, headers=self.default_headers, json=params)

  def index(self, **query_params: EndpointsIndexQueryParams) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def show(self, endpoint_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)

  def destroy(self, endpoint_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.ENDPOINTS_ENDPOINT}/{endpoint_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.delete(url, headers=self.default_headers, params=query_params)


from typing import List, TypedDict

from .pagination_query_params import PaginationQueryParams

class ScenarioShowResponse(TypedDict):
  created_at: str
  description: str
  key: str
  name: str
  overwritable: bool
  updated_at: str

class ScenariosIndexQueryParams(PaginationQueryParams):
  pass

class ScenariosIndexResponse(TypedDict):
  list: List[ScenarioShowResponse]
  total: int



from stoobly_orator.orm import belongs_to

from stoobly_agent.app.models.schemas.request import Request

from .base import Base

class Response(Base):
  __fillable__ = ['control', 'http_version', 'request_id', 'raw']

  @belongs_to
  def request(self):
    return Request


import pdb

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow

from stoobly_agent.app.proxy.record.joined_request import JoinedRequest
from stoobly_agent.lib.orm.types.request_columns import RequestColumns
from stoobly_agent.lib.orm.types.response_columns import ResponseColumns

from ..orm_request_builder import ORMRequestBuilder

def build_request_columns(flow: MitmproxyHTTPFlow, joined_request: JoinedRequest, **params):
  builder = ORMRequestBuilder()
  request_columns = builder.columns_from_mitmproxy_request(flow.request)
  response_columns = builder.columns_from_mitmproxy_response(flow.response)

  request_columns: RequestColumns = {
    **request_columns,
    **response_columns,
    'control': joined_request.request_string.control,
    'latency': joined_request.response_string.latency,
    'raw': joined_request.request_string.get(),
    'status': flow.response.status_code, 
  }

  return { 
    **request_columns, 
    **params, 
    **{ 'uuid': params.get('uuid') or joined_request.request_string.request_id.strip() } 
  }

def build_response_columns(flow: MitmproxyHTTPFlow, joined_request: JoinedRequest):
  response_columns: ResponseColumns = {
    'control': joined_request.response_string.control,
    'http_version': __http_version(flow.response.http_version),
    'raw': joined_request.response_string.get(),
  }

  return response_columns

def __http_version(http_version: str):
  if not isinstance(http_version, str):
    return http_version
  _version = http_version.split('/')[1]
  return float(_version)


from typing import TypedDict

class Header(TypedDict):
  name: bytes
  value: bytes

class Request(TypedDict):
  body: bytes
  headers: dict
  method: str
  status: bytes
  status_code: int
  url: bytes

class RequestParseHandler():
  __request_shell: Request = None

  def __init__(self, request_shell: dict = {}):
    self.__request_shell = request_shell
    self.__request_shell['headers'] = {}
    self.__request_shell['status_code'] = None
    self.__request_shell['method'] = None
    self.__request_shell['body'] = b''

  def on_body(self, body: bytes):
    self.__request_shell['body'] = body

  def on_header(self, name: bytes, value: bytes):
    self.__request_shell['headers'][name] = value

  def on_status(self, status: bytes):
    self.__request_shell['status'] = status

  def on_url(self, url: bytes):
    self.__request_shell['url'] = url


import os
import pdb

from typing import TypedDict

from .app import App
from .app_command import AppCommand

class AppCreateOptions(TypedDict):
  name: str
  network: str

class AppCreateCommand(AppCommand):

    def __init__(self, app: App, **kwargs: AppCreateOptions):
        super().__init__(app)

        if kwargs.get('app_name'):
            self.app_config.name = kwargs['app_name']

        if kwargs.get('network'):
            self.app_config.network = kwargs['network']

    @property
    def app_name(self):
        return self.app_config.name

    @property
    def app_network(self):
        return self.app_config.network

    def build(self):
        dest = self.scaffold_namespace_path

        self.app.copy_folders_and_hidden_files(self.app_templates_root_dir, dest)

        with open(os.path.join(dest, '.gitignore'), 'w') as fp:
            fp.write("\n".join(['gateway/.docker-compose.base.yml', '**/.env']))

        self.app_config.write()


import os
import pdb
import pytest

from click.testing import CliRunner
from mock import patch
from unittest.mock import MagicMock

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

# Enable remote feature
from stoobly_agent.config.constants import env_vars, test_strategy
os.environ[env_vars.FEATURE_REMOTE] = '1'

from stoobly_agent.cli import record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestScenarioTestIntegration():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenDiffTest():
        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def diff_matches(self, runner: CliRunner, created_scenario: Scenario):
            with patch('stoobly_agent.app.proxy.test.test_service.diff_matches') as spy:
                spy.return_value = (True, '')
                test_result = runner.invoke(scenario, ['test', created_scenario.key()])
                assert test_result.exit_code == 0

                return spy

        def test_it_calls_diff_matches_once(self, diff_matches: MagicMock):
            assert diff_matches.call_count == 1

    class TestWhenFuzzyTest():
        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def fuzzy_matches(self, runner: CliRunner, created_scenario: Scenario):
            with patch('stoobly_agent.app.proxy.test.test_service.fuzzy_matches') as spy:
                spy.return_value = (True, '')
                test_result = runner.invoke(scenario, ['test', '--strategy', test_strategy.FUZZY, created_scenario.key()])
                assert test_result.exit_code == 0

                return spy

        def test_it_calls_fuzzy_matches_once(self, fuzzy_matches: MagicMock):
            assert fuzzy_matches.call_count == 1

    class TestWhenCustomTest():
        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def custom_matches(self, runner: CliRunner, created_scenario: Scenario):
            with patch('stoobly_agent.app.proxy.test.test_service.custom_matches') as spy:
                spy.return_value = (True, '')
                test_result = runner.invoke(scenario, ['test', '--strategy', test_strategy.CUSTOM, created_scenario.key()])
                assert test_result.exit_code == 0

                return spy

        def test_it_calls_custom_matches_once(self, custom_matches: MagicMock):
            assert custom_matches.call_count == 1

    class TestWhenContractTest():
        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def contract_matches(self, runner: CliRunner, created_scenario: Scenario):
            with patch('stoobly_agent.app.proxy.test.test_service.contract_matches') as spy:
                spy.return_value = (True, '')
                test_result = runner.invoke(scenario, ['test', '--strategy', test_strategy.CONTRACT, created_scenario.key()])
                assert test_result.exit_code == 0

                return spy

        def test_it_calls_contraact_matches_once(self, contract_matches: MagicMock):
            assert contract_matches.call_count == 1


import pdb
import requests

from typing import Tuple

from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory
from stoobly_agent.app.proxy.mock.request_hasher import RequestHasher
from stoobly_agent.app.proxy.record.response_string_control import ResponseStringControl
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.lib.orm.transformers import ORMToRequestsResponseTransformer
from stoobly_agent.lib.utils.decode import decode

from .local_db_adapter import LocalDBAdapter

class LocalDBResponseAdapter(LocalDBAdapter):
  __request_orm = None

  def __init__(self, request_orm: Request.__class__ = Request):
    self.__request_orm = request_orm

  def show(self, request_id):
    request = self.__request_orm.find(request_id)

    if not request:
      return self.__request_not_found()

    response = request.response
    return self.success([self.__to_show_response(response)])

  def mock(self, request_id) -> Tuple[requests.Response, int]:
    request = self.__request_orm.find(request_id)

    if not request:
      return self.__request_not_found()

    response = request.response
    return self.success(ORMToRequestsResponseTransformer(response).with_response_id().transform())

  def update(self, response_id, **params):
    response = Response.find(response_id)

    if not response:
      return self.__request_not_found()
    
    transformer = ORMToRequestsResponseTransformer(response)

    request_params = {}

    if params.get('status'):
      transformer.with_status(params['status'])

    headers = params.get('headers') or {}
    if params.get('headers'):
      transformer.with_headers(headers)
      request_params['response_headers_hash'] = RequestHasher.instance().hash_params(headers)
 
    # If text not set and content-encoding changed 
    current_python_response = ORMToRequestsResponseTransformer(response).transform()
    if not params.get('text'):
      if headers.get('content-encoding') != current_python_response.headers.get('content-encoding'):
        params['text'] = current_python_response.content

    if params.get('text'):
      text: str = params['text'] # This will be given in decoded format
      content_encoding = self.content_encoding_header(headers, current_python_response.headers)
      encoded_text = self.encode_body(text, content_encoding)

      transformer.with_body(encoded_text)
      request_params['response_hash'] = RequestHasher.instance().hash_text(encoded_text)
      
    if params.get('latency'):
      control = response.control
      control = ResponseStringControl(control)
      control.timestamp = control.timestamp - control.latency * 1000 + int(params['latency']) * 1000
      control.latency = params['latency']
      params['control'] = control.serialize()
      del params['latency']

    if transformer.dirty:
      python_response = transformer.transform()
      params = {
        **params,
        'raw': PythonResponseAdapterFactory(python_response).raw_response()
      }

    if response.update(params):
      if len(request_params.keys()) != 0:
        request = self.__request_orm.find(response.request_id)

        if request:
          request.update(request_params)

      return self.success(self.__to_show_response(response))

    return self.internal_error('Could not update response')

  def __to_show_response(self, response: Response):
    python_response = ORMToRequestsResponseTransformer(response).transform()
    headers = python_response.headers
    content = python_response.content

    return {
      'id': response.id,
      'mime_type': headers.get('content-type'),
      'text': decode(content),
    }

  def __request_not_found(self):
    return self.not_found('Request not found')


import pdb
import requests

from typing import Tuple

from stoobly_agent.app.models.types import ReplayedResponseIndexQueryParams
from stoobly_agent.lib.orm.replayed_response import ReplayedResponse
from stoobly_agent.lib.orm.transformers.orm_to_requests_response_transformer import ORMToRequestsResponseTransformer

from .local_db_adapter import LocalDBAdapter

class LocalDBReplayedResponseAdapter(LocalDBAdapter):
  __replayed_response_orm = None

  def __init__(self, replayed_response_orm: ReplayedResponse.__class__ = ReplayedResponse):
    self.__replayed_response_orm = replayed_response_orm

  def create(self, **body_params):
    replayed_response = ReplayedResponse.create(**body_params)
    return self.success(self.__transform_replayed_response(replayed_response))

  def index(self, **query_params: ReplayedResponseIndexQueryParams):
    request_id = int(query_params.get('request_id')) if query_params.get('request_id') else None
    page = int(query_params.get('page') or 0)
    size = int(query_params.get('size') or 20)
    sort_by = query_params.get('sort_by') or 'id'
    sort_order = query_params.get('sort_order') or 'desc'

    replayed_responses = self.__replayed_response_orm.where('request_id', request_id)

    total = replayed_responses.count()
    replayed_responses = replayed_responses.offset(page * size).limit(size).order_by(sort_by, sort_order).get()

    return self.success({
      'list': list(map(lambda replayed_response: self.__transform_replayed_response(replayed_response), replayed_responses)),
      'total': total,
    })

  def destroy(self, replayed_response_id: int):
    replayed_response = self.__replayed_response_orm.find(replayed_response_id)

    if not replayed_response:
      return self.__replayed_response_not_found()

    replayed_response.delete()

    return self.success({})

  def mock(self, replayed_response_id: int) -> Tuple[requests.Response, int]:
    replayed_response = self.__replayed_response_orm.find(replayed_response_id)

    if not replayed_response:
      return self.__replayed_response_not_found()

    return self.success(ORMToRequestsResponseTransformer(replayed_response).transform())

  def raw(self, replayed_response_id: int):
    replayed_response = self.__replayed_response_orm.find(replayed_response_id)

    if not replayed_response:
      return self.__replayed_response_not_found()

    return self.success(replayed_response.raw)

  def activate(self, replayed_response_id: int):
    replayed_response: ReplayedResponse = self.__replayed_response_orm.find(replayed_response_id)

    if not replayed_response:
      return self.__replayed_response_not_found()

    request = replayed_response.request
    response = request.response

    _replayed_response = ReplayedResponse.create(
      latency=request.latency,
      raw=response.raw,
      request_id=request.id,
      status=request.status,
    )
    _replayed_response.created_at = request.created_at
    _replayed_response.save()

    request.created_at = replayed_response.created_at
    request.latency = replayed_response.latency
    request.status = replayed_response.status
    request.save()

    response.created_at = replayed_response.created_at
    response.raw = replayed_response.raw
    response.save()

    replayed_response.delete()

    return self.success(self.__transform_replayed_response(_replayed_response))
    
  def __transform_replayed_response(self, replayed_response: ReplayedResponse):
    res = replayed_response.to_dict()
    res.pop('raw')
    return res

  def __replayed_response_not_found(self):
    return self.not_found('Replayed response not found')


import pdb
from typing import TypedDict

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from requests import Response

from stoobly_agent.lib.api.tests_resource import TestsResource
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.logger import Logger, bcolors

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings

from ...intercept_settings import InterceptSettings
from ...record.join_request_service import join_request_from_flow

LOG_ID = 'Test'

class UploadTestData(TypedDict):
  expected_response: str
  log: str
  passed: bool
  request_id: int
  status: int
  strategy: str

def inject_upload_test(
  api: TestsResource,
  intercept_settings: InterceptSettings, 
):
  settings = intercept_settings.settings

  if not api:
    api = TestsResource(settings.remote.api_url, settings.remote.api_key)

  if not intercept_settings:
    intercept_settings = InterceptSettings(Settings.instance(LOG_ID))

  return lambda flow, **kwargs: upload_test(api, intercept_settings, flow, **kwargs)

def upload_test(
  api: TestsResource,
  intercept_settings: InterceptSettings, 
  flow: MitmproxyHTTPFlow, 
  **kwargs: UploadTestData
) -> Response:
    joined_request = join_request_from_flow(flow, intercept_settings)

    Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Uploading{bcolors.ENDC} test results for {joined_request.proxy_request.url()}")

    raw_requests = joined_request.build()

    # If report key is set, upload test to report
    report_key = intercept_settings.report_key
    if report_key:
      Logger.instance(LOG_ID).debug(f"Using report {report_key}")

      api.with_report_key(report_key, kwargs)

    scenario_key = intercept_settings.scenario_key
    if scenario_key:
      return api.from_scenario_key(
        scenario_key, 
        lambda project_id, query_params: api.create(project_id, raw_requests, { **query_params, **kwargs })
      )
    else:
      return api.from_project_key(
        intercept_settings.project_key,
        lambda project_id: api.create(project_id, raw_requests, { **kwargs })
      )


import pdb

from stoobly_agent.app.models.factories.resource.request import RequestResourceFactory
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.api.header_names_resource import HeaderNamesResource
from stoobly_agent.lib.api.body_param_names_resource import BodyParamNamesResource
from stoobly_agent.lib.api.query_param_names_resource import QueryParamNamesResource
from stoobly_agent.lib.api.response_param_names_resource import ResponseParamNamesResource
from stoobly_agent.lib.api.response_header_names_resource import ResponseHeaderNamesResource

from .endpoints_apply_context import EndpointsApplyContext
from .endpoints_import_context import EndpointsImportContext
from .endpoints_apply_service import apply_endpoints
from .endpoints_import_service import import_endpoints
from .synchronize_request_service import SynchronizeRequestService

class EndpointFacade():
  def __init__(self, __settings: Settings):
    self.__settings = __settings
  
  def apply(self, context: EndpointsApplyContext):
    local_db_request_adapter = RequestResourceFactory(self.__settings.remote).local_db()
    context.with_request_model(local_db_request_adapter)

    lifecycle_hooks = context.lifecycle_hooks
    synchronize_request_service = SynchronizeRequestService(local_db_request_adapter=local_db_request_adapter)
    request_handler = lambda request, endpoint: synchronize_request_service.synchronize_request(
      request, endpoint, lifecycle_hooks
    )
    context.with_request_handler(request_handler)

    apply_endpoints(context)
  
  def import_endpoints(self, context: EndpointsImportContext):
    api_url = self.__settings.remote.api_url
    api_key = self.__settings.remote.api_key

    context.with_endpoint_resource(EndpointsResource(api_url, api_key))
    context.with_header_name_resource(HeaderNamesResource(api_url, api_key))
    context.with_body_param_name_resource(BodyParamNamesResource(api_url, api_key))
    context.with_query_param_name_resource(QueryParamNamesResource(api_url, api_key))
    context.with_response_param_name_resource(ResponseParamNamesResource(api_url, api_key))
    context.with_response_header_name_resource(ResponseHeaderNamesResource(api_url, api_key))

    import_endpoints(context)


from stoobly_orator.migrations import Migration

class CreateTraceAliases(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('trace_aliases') as table:
            table.increments('id')
            table.integer('trace_id').unsigned()
            table.foreign('trace_id').references('id').on('traces').index()
            table.text('name')
            table.binary('value')
            table.string('value_inferred_type').nullable()
            table.text('assigned_to').nullable()
            table.string('assigned_to_inferred_type').nullable()
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('trace_aliases')



import os
import pdb
import yaml

from stoobly_agent.lib.logger import Logger

from .app import App
from .constants import BIN_FOLDER_NAME, COMPOSE_TEMPLATE, CONFIG_FILE, ENV_FILE, PUBLIC_FOLDER_NAME
from .docker.constants import DOCKER_COMPOSE_CUSTOM
from .service_command import ServiceCommand
from .workflow_config import WorkflowConfig

LOG_ID = 'WorkflowCommand'

class WorkflowCommand(ServiceCommand):

  def __init__(self, app: App, **kwargs):
    super().__init__(app, **kwargs)

    self.__workflow_name = kwargs['workflow_name']
    self.__config = WorkflowConfig(self.workflow_path, **kwargs)

  @property
  def bin_dir_path(self):
    return os.path.join(self.workflow_path, BIN_FOLDER_NAME)

  @property
  def compose_path(self):
    return os.path.join(
      self.data_dir_path,
      self.compose_relative_path
    )

  @property
  def compose_relative_path(self):
    return os.path.join(
      self.workflow_relative_path, 
      COMPOSE_TEMPLATE.format(workflow=self.workflow_name)
    )

  @property
  def containers(self):
    if not os.path.exists(self.compose_path):
      return []

    with open(self.compose_path, 'r') as fp:
      config = yaml.safe_load(fp)
      services = config.get('services')
      if not isinstance(services, dict):
        return []

      return services
        
  @property
  def custom_compose(self):
    if os.path.exists(self.custom_compose_path):
      with open(self.custom_compose_path, 'r') as fp:
        compose = yaml.safe_load(fp)
        if compose and not isinstance(compose, dict):
          Logger.instance(LOG_ID).warn(f"Could not parse {self.custom_compose_path} as YAML")
          return {}
        return compose

  @property
  def custom_compose_relative_path(self):
    return os.path.join(
      self.workflow_relative_path,
      DOCKER_COMPOSE_CUSTOM
    )

  @property
  def custom_compose_path(self):
    return os.path.join(
      self.data_dir_path,
      self.custom_compose_relative_path
    )

  @property
  def custom_services(self):
    custom_compose = self.custom_compose

    if not custom_compose:
      return {}
    
    services = custom_compose.get('services')
    if not isinstance(services, dict):
      return {}

    return services

  @property
  def public_dir_path(self):
    return os.path.join(self.workflow_path, PUBLIC_FOLDER_NAME)

  @property
  def workflow_config(self):
    return self.__config

  @property
  def workflow_config_path(self):
    return os.path.join(self.workflow_path, CONFIG_FILE)

  @property
  def workflow_env_path(self):
    return os.path.join(self.workflow_path, ENV_FILE)

  @property
  def workflow_exists(self):
    return os.path.exists(self.workflow_path)

  @property
  def workflow_name(self):
    return self.__workflow_name

  @property
  def workflow_path(self):
    return os.path.join(
      self.data_dir_path,
      self.workflow_relative_path
    )

  @property
  def workflow_relative_path(self):
    return os.path.join(
      self.service_relative_path,
      self.workflow_name
    )

  @property
  def workflow_templates_root_dir(self):
    return os.path.join(self.templates_root_dir, 'workflow')
  
  def env(self, _c: dict):
    _config = self.app_config.read()
    _config.update(self.service_config)
    _config.update(self.workflow_config)
    _config.update(_c)
    return _config



import json
import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.cli.helpers.print_service import JSON_FORMAT
from stoobly_agent.cli import record, request
from stoobly_agent.lib.api.keys import RequestKey
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
      return CliRunner()

class TestList():

  @pytest.fixture(autouse=True)
  def settings(self):
    return reset()

  def test_it_shows_request(self, runner: CliRunner):
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)

    request_result = runner.invoke(request, ['list', '--without-headers', '--select', 'key'])
    assert request_result.exit_code == 0

    output = request_result.stdout
    request_key = RequestKey(output.strip())

    _request = Request.find_by(uuid=request_key.id)

    assert _request.url == DETERMINISTIC_GET_REQUEST_URL

  def test_it_shows_all_requests(self, runner: CliRunner):
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)

    request_result = runner.invoke(request, ['list', '--without-headers', '--select', 'key'])
    assert request_result.exit_code == 0

    output = request_result.stdout
    assert len(output.strip().split("\n")) == 2

  def test_it_shows_requests_on_page(self, runner: CliRunner):
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)

    request_result = runner.invoke(request, ['list', '--without-headers', '--select', 'key', '--page', '2', '--size', '1'])
    assert request_result.exit_code == 0

    output = request_result.stdout
    assert len(output.strip().split("\n")) == 1

  def test_it_sorts_by_host(self, runner: CliRunner):
    self.__record_request(runner, 'https://www.google.com')
    self.__record_request(runner, 'https://www.facebook.com')

    request_result = runner.invoke(request, ['list', '--without-headers', '--select', 'host', '--sort-by', 'host', '--sort-order', 'asc'])
    assert request_result.exit_code == 0

    output = request_result.stdout
    host = output.strip().split("\n")[0]
    assert host == 'www.facebook.com'

  def test_it_formats_json(self, runner: CliRunner):
    self.__record_request(runner, DETERMINISTIC_GET_REQUEST_URL)

    request_result = runner.invoke(request, ['list', '--format', JSON_FORMAT])
    assert request_result.exit_code == 0

    output = request_result.stdout
    requests = json.loads(output)

    assert requests[0].get('status') == '200'

  def __record_request(self, runner: CliRunner, url: str):
    record_result = runner.invoke(record, [url])
    assert record_result.exit_code == 0


import json
import pdb
import urllib.parse

from email.message import Message
from mitmproxy.coretypes.multidict import MultiDict
from mitmproxy.net import encoding
from typing import Dict, Union

from stoobly_agent.lib.utils.decode import decode

from .multipart import decode as multipart_decode, encode as multipart_encode

JSON = 'application/json'
MULTIPART_FORM = 'multipart/form-data'
WWW_FORM_URLENCODED = 'application/x-www-form-urlencoded'

def compress(body: Union[bytes, str], content_encoding: Union[None, str]) -> Union[bytes, str]:
    if content_encoding:
      return encoding.encode(body, content_encoding)
    else:
      return body

def decode_response(content: Union[bytes, str], content_type: Union[None, str]) -> Union[dict, list, MultiDict]:
    if not content_type:
        return content

    _content_type = normalize_header(content_type)

    decoded_response = content
    if is_json(_content_type):
        content = decode(content)
        decoded_response = parse_json(content)
    elif _content_type == WWW_FORM_URLENCODED:
        decoded_response = parse_www_form_urlencoded(content)
    elif _content_type == MULTIPART_FORM:
        decoded_response = parse_multipart_form_data(content, content_type)

    return decoded_response

def decompress(body: Union[bytes, str], content_encoding: Union[None, str]) -> Union[bytes, str]:
    if content_encoding:
      return encoding.decode(body, content_encoding)
    else:
      return body

def encode_response(content, content_type: Union[bytes, None, str]) -> Union[bytes, str]:
    if not content_type:
        #raise ValueError('Missing content_type value')
        return content

    _content_type = normalize_header(content_type)

    encoded_response = content
    if is_json(_content_type):
        encoded_response = serialize_json(content)
    elif _content_type == WWW_FORM_URLENCODED:
        encoded_response = serialize_www_form_urlencoded(content)
    elif _content_type == MULTIPART_FORM:
        encoded_response = serialize_multipart_form_data(content, content_type)

    return encoded_response   

def parse_json(content):
    try:
        return json.loads(content)
    except:
        return content

def parse_multipart_form_data(content, content_type) -> Dict[bytes, bytes]:
    headers = {'content-type': content_type}
    decoded_multipart = multipart_decode(headers, content)

    if not decoded_multipart:
        return content
 
    return MultiDict(decoded_multipart)

def parse_www_form_urlencoded(content):
    try:
        return urllib.parse.parse_qs(content)
    except:
        return content

def serialize_json(o):
    return json.dumps(o)

def serialize_multipart_form_data(o: MultiDict, content_type: Union[bytes, str]) -> bytes:
    _o = MultiDict()
    for k, v in o.items():
        if isinstance(k, str):
            k = k.encode()

        if not isinstance(v, str) and not isinstance(v, bytes):
            v = str(v)

        if isinstance(v, str):
            v = v.encode()
        
        _o.add(k, v)

    headers = {'content-type': content_type}
    return multipart_encode(headers, _o.items())

def serialize_www_form_urlencoded(o):
    return urllib.parse.urlencode(o)

def normalize_header(header):
    if isinstance(header, bytes):
        header = header.decode('utf-8')
    return __parse_separated_header(header).lower()

def is_traversable(content):
    return isinstance(content, list) or isinstance(content, dict) or isinstance(content, MultiDict)

def is_json(content_type):
   _content_type = content_type.lower()
   return _content_type == JSON or _content_type.startswith('application/x-amz-json') 


def __parse_separated_header(header: str):
    # Adapted from https://peps.python.org/pep-0594/#cgi
    message = Message()
    message['content-type'] = header 
    return message.get_content_type()




import inspect
import os
import pdb
import subprocess

from typing import TypedDict

from stoobly_agent import COMMAND
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.config.constants.env_vars import ENV
from stoobly_agent.app.settings.constants.mode import TEST

class MockOptions(TypedDict):
  data: str
  headers: dict
  method: str
  project_key: str
  scenario_key: str

class Mock():

  @classmethod
  def get(cls, url: str, **options: MockOptions):
    return cls.mock(url, **{
      **options,
      'method': 'GET',
    }) 

  @classmethod
  def post(cls, url: str, **options: MockOptions):
    return cls.mock(url, **{
      **options,
      'method': 'POST',
    })

  @classmethod
  def put(cls, url: str, **options: MockOptions):
    return cls.mock(url, **{
      **options,
      'method': 'PUT',
    })

  @classmethod
  def delete(cls, url: str, **options: MockOptions):
    return cls.mock(url, **{
      **options,
      'method': 'DELETE',
    })

  @staticmethod
  def mock(url: str, **options: MockOptions):
    command = [COMMAND, 'mock', '--format', 'raw']

    if isinstance(options.get('data'), str):
      data_option = f"--data {options['data']}"
      command.append(data_option)

    if isinstance(options.get('headers'), dict):
      options = []
      for k, v in options['headers'].items:
        options.append('--header')
        options.append(f"{k}: {v}")

      if len(options) > 0:
        command.append(' '.join(options))

    if isinstance(options.get('project_key'), str):
      command.append('--project_key')
      command.append(options['project_key'])

    if isinstance(options.get('method'), str):
      command.append('--request')
      command.append(f"{options['method']}")      

    if isinstance(options.get('scenario_key'), str):
      command.append('--scenario_key')
      command.append(options['scenario_key'])

    command.append(url)

    # We do not want the child process to run in test mode
    if os.environ.get(ENV) == TEST:
      del os.environ[ENV]

    # Change dir to caller file's directory
    dir_path = inspect.stack()[2].filename # Path of calling method
    parent_dir = os.path.dirname(dir_path)
    if len(parent_dir) > 0:
      os.chdir(parent_dir)

    completed_process = subprocess.run(command, check=True, stdout=subprocess.PIPE)

    if completed_process.returncode == 0:
      adapter = RawHttpResponseAdapter(completed_process.stdout) 
      response = adapter.to_response()
      return response


import subprocess
import sys

from dotenv import load_dotenv

def exec_stream(command):
  load_dotenv()

  # Start the process
  process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
  
  # Stream output line by line
  while True:
      stdout = process.stdout.readline()
      stderr = process.stderr.readline()

      if stdout == '' and stderr == '' and process.poll() is not None:
          break

      if stdout:
          print(stdout, end='')

      if stderr:
          print(stderr, end='', file=sys.stderr)
  
  # Get the remaining output (if any)
  rc = process.poll()
  return rc


import asyncio
import os
import pdb
import signal

from mitmproxy.net import tls

from stoobly_agent.app.cli.scaffold.constants import SERVICE_NAME_ENV
from stoobly_agent.app.cli.scaffold.templates.constants import CORE_MOCK_UI_SERVICE_NAME
from stoobly_agent.config.mitmproxy import MitmproxyConfig

# Monkey patch for OpenSSL unsafe legacy renegotiation disabled
# See: https://stackoverflow.com/questions/71603314/ssl-error-unsafe-legacy-renegotiation-disabled
tls.DEFAULT_OPTIONS |= 0x4 

from mitmproxy.options import Options
from mitmproxy.tools.dump import DumpMaster

from stoobly_agent.app.settings import Settings

INTERCEPT_HANDLER_FILENAME = 'intercept_handler.py'

def run(**kwargs):
    async def main():
        options = Options()
        master = DumpMaster(options)
        config = MitmproxyConfig.instance() 
        config.with_master(master) # Initialize mitmproxy config instance

        cli_options = kwargs.copy()
        __with_static_options(config, cli_options)
        __with_cli_options(config, cli_options)

        config.dump()

        loop = asyncio.get_running_loop()

        def _sigint(*_):
            loop.call_soon_threadsafe(getattr(master, "prompt_for_exit", master.shutdown))

        def _sigterm(*_):
            loop.call_soon_threadsafe(master.shutdown)

        # We can't use loop.add_signal_handler because that's not available on Windows' Proactorloop,
        # but signal.signal just works fine for our purposes.
        signal.signal(signal.SIGINT, _sigint)
        signal.signal(signal.SIGTERM, _sigterm)

        await master.run()

        return master

    return asyncio.run(main())

def __get_intercept_handler_path():
    cwd = os.path.dirname(os.path.realpath(__file__))
    script = os.path.join(cwd, INTERCEPT_HANDLER_FILENAME)
    return script

def __with_static_options(config: MitmproxyConfig, cli_options):
    options = (
        'block_global=false',
        f"scripts={__get_intercept_handler_path()}",
        'upstream_cert=false'
    )

    config.set(options)

def __with_cli_options(config: MitmproxyConfig, cli_options: dict):
    __commit_options(cli_options)
    __filter_options(cli_options)

    options = []
    def append_option(key, val):
        if isinstance(val, bool):
            options.append(f"{key}={str(val).lower()}")
        else:
            options.append(f"{key}={val}")

    for key, val in cli_options.items():
        if isinstance(val, tuple):
            for v in val:
                append_option(key, v)
        else:
            append_option(key, val)

    config.set(tuple(options))

def __commit_options(options: dict):
    changed = False
    service_name = os.environ.get(SERVICE_NAME_ENV)
    settings = Settings.instance()

    if not service_name or options.get('intercept'):
        # In the case when service name is set,
        # only update if intercept is explicitly enabled
        intercept = not not options.get('intercept')
        changed = settings.proxy.intercept.active != intercept
        settings.proxy.intercept.active = intercept

    if not service_name or service_name == CORE_MOCK_UI_SERVICE_NAME:
        # Causes potentially unintended side effects when run as part of scaffold
        # Defer to ui service for configuration in this case
        if options.get('proxy_host') and options.get('proxy_port'):
            settings.proxy.url = f"http://{options.get('proxy_host')}:{options.get('proxy_port')}"

        settings.ui.active = not options.get('headless')

        changed = True

    if changed:
        settings.commit()

def __filter_options(options):
    ''' 
    Filter out non-mitmproxy options
    '''

    options['confdir'] = options['ca_certs_dir_path']
    options['listen_host'] = options['proxy_host']
    options['listen_port'] = options['proxy_port']
    options['mode'] = options['proxy_mode']

    if 'api_url' in options:
        del options['api_url']

    if 'ca_certs_dir_path' in options:
        del options['ca_certs_dir_path']

    if 'certs' in options and not options['certs']:
        del options['certs']

    if 'cert_passphrase' in options and not options['cert_passphrase']:
        del options['cert_passphrase']

    if 'headless' in options:
        del options['headless']

    if 'intercept' in options:
        del options['intercept']

    if 'lifecycle_hooks_path' in options:
        del options['lifecycle_hooks_path']

    if 'proxyless' in options:
        del options['proxyless']

    if 'public_directory_path' in options:
        del options['public_directory_path']

    if 'response_fixtures_path' in options:
        del options['response_fixtures_path']

    if 'ui_host' in options:
        del options['ui_host']

    if 'ui_port' in options:
        del options['ui_port']

    del options['log_level']
    del options['proxy_host']
    del options['proxy_mode']
    del options['proxy_port']


import requests

from stoobly_agent.lib.api.reports_resource import ReportsResource
from stoobly_agent.app.settings import Settings

class ReportFacade():

  def __init__(self, settings: Settings):
    self.settings = settings

  def create(self, project_key: str, name: str, description: str = ''):
    api = ReportsResource(self.settings.remote.api_url, self.settings.remote.api_key)

    res: requests.Response = api.from_project_key(
      project_key, 
      lambda project_id: api.create(
        project_id, {
          'description': description,
          'name': name,
        }
      )
    )

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()

  def index(self, project_key: str, kwargs: dict):
    api = ReportsResource(self.settings.remote.api_url, self.settings.remote.api_key)

    res: requests.Response = api.from_project_key(
      project_key, 
      lambda project_id: api.index(
        project_id, kwargs
      )
    )

    if not res.ok:
      raise AssertionError(res.content)

    return res.json()



import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.models.adapters.orm import JoinedRequestStringAdapter
from stoobly_agent.app.models.factories.resource.local_db.helpers.log import Log
from stoobly_agent.app.models.factories.resource.local_db.helpers.log_event import DELETE_ACTION
from stoobly_agent.app.models.factories.resource.local_db.helpers.request_snapshot import RequestSnapshot
from stoobly_agent.cli import record, request
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestRequestSnapshot():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
      return reset()

    class TestWhenReplacing():

      @pytest.fixture(scope='class')
      def recorded_request(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      def test_it_snapshots(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        snapshot = RequestSnapshot(recorded_request.uuid)
        raw_request = snapshot.request

        expected_raw_request = JoinedRequestStringAdapter(recorded_request).adapt()

        assert raw_request == expected_raw_request

        log = Log()

        events = log.raw_events
        assert len(events) == 1

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 1

      def test_it_replaces(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request.key()])
        assert snapshot_result.exit_code == 0

        snapshot = RequestSnapshot(recorded_request.uuid)
        raw_request = snapshot.request

        expected_raw_request = JoinedRequestStringAdapter(recorded_request).adapt()

        assert raw_request == expected_raw_request

        log = Log()

        events = log.raw_events
        assert len(events) == 2

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 1

      def test_it_deletes(self, runner: CliRunner, recorded_request: Request):
        snapshot_result = runner.invoke(request, ['snapshot', '--action', DELETE_ACTION, recorded_request.key()])
        assert snapshot_result.exit_code == 0

        log = Log()

        events = log.raw_events
        assert len(events) == 3

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 0

    class TestWhenAppending():

      @pytest.fixture(scope='class')
      def recorded_request_one(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      @pytest.fixture(scope='class')
      def recorded_request_two(self, runner: CliRunner):
        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

      def test_it_appends(self, runner: CliRunner, recorded_request_one: Request, recorded_request_two: Request):
        snapshot_result = runner.invoke(request, ['snapshot', recorded_request_one.key()])
        assert snapshot_result.exit_code == 0

        snapshot_result = runner.invoke(request, ['snapshot', recorded_request_two.key()])
        assert snapshot_result.exit_code == 0

        log = Log()

        events = log.raw_events
        assert len(events) == 2

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 2

      def test_it_deletes(self, runner: CliRunner, recorded_request_one: Request, recorded_request_two: Request):
        snapshot_result = runner.invoke(request, ['snapshot', '--action', DELETE_ACTION, recorded_request_one.key()])
        assert snapshot_result.exit_code == 0

        log = Log()

        events = log.raw_events
        assert len(events) == 3

        unprocessed_events = log.unprocessed_events
        assert len(unprocessed_events) == 1

        event = unprocessed_events[0]
        assert event.resource_uuid == recorded_request_two.uuid



import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.settings import Settings
from stoobly_agent.app.settings.constants import firewall_action
from stoobly_agent.cli import config
from stoobly_agent.config.constants import mode
from stoobly_agent.lib.api.keys.project_key import LOCAL_PROJECT_ID

@pytest.fixture
def settings():
  return reset()

class TestFirewall():

  def test_it_sets_rule(self, settings: Settings):
    runner = CliRunner()

    method = 'POST'
    _mode = mode.MOCK
    pattern = '.*?'

    firewall_result = runner.invoke(config, [
        'firewall', 'set', 
        '--method', method, '--mode', _mode, '--pattern', pattern, '--action', firewall_action.EXCLUDE
      ]
    )
    assert firewall_result.exit_code == 0

    settings.load()

    firewall_rules = settings.proxy.firewall.firewall_rules(str(LOCAL_PROJECT_ID))
    assert len(firewall_rules) == 1

    firewall_rule = firewall_rules[0]
    assert firewall_rule.pattern == pattern
    assert firewall_rule.methods == [method]
    assert firewall_rule.action == firewall_action.EXCLUDE
    assert firewall_rule.modes == [_mode]


from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.api.requests_resource import RequestsResource
from stoobly_agent.app.settings import RemoteSettings

from .local_db.request_adapter import LocalDBRequestAdapter
from .stoobly.request_adapter import StooblyRequestAdapter

class RequestResourceFactory():

  def __init__(self, settings: RemoteSettings):
    self.__remote_settings = settings

  def local_db(self) -> LocalDBRequestAdapter:
    return LocalDBRequestAdapter(Request)  

  def stoobly(self) -> StooblyRequestAdapter:
    api = RequestsResource(self.__remote_settings.api_url, self.__remote_settings.api_key)
    return StooblyRequestAdapter(api)


import pdb
import requests

from stoobly_orator.migrations import Migration
from urllib.parse import urlparse

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.proxy.record.response_string_control import ResponseStringControl
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.utils.decode import decode

class AlignRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('requests') as table:
            table.integer('latency').default(0)
            table.string('password').nullable()
            table.string('query').nullable()
            table.integer('status').nullable()
            table.string('user').nullable()

        for request in Request.all():
            _request: requests.Request = RawHttpRequestAdapter(request.raw)

            parsed_url = urlparse(_request.url)
            request.path = parsed_url.path
            request.query = parsed_url.query

            if parsed_url.username:
                request.user = parsed_url.username

            if parsed_url.password:
                request.password = parsed_url.password

            response = request.response
            if response:
                response_control = ResponseStringControl()
                response_control.parse(decode(response.control))

                request.latency = response_control.latency

                _response = RawHttpResponseAdapter(response.raw).to_response()
                request.status = int(_response.status_code)

            request.update()

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('requests') as table:
            table.drop_column('latency', 'password', 'query', 'status', 'user')



from typing import Callable, TypedDict, Union

class PrintOptions(TypedDict):
  filter: list
  print_handler: Union[Callable, None]
  select: list

class TabulatePrintOptions(PrintOptions):
  format: str
  headers: bool


import requests

from stoobly_agent.app.models.adapters.python import PythonResponseAdapterFactory
from stoobly_agent.app.proxy.mitmproxy.response_facade import MitmproxyResponseFacade
from stoobly_agent.app.proxy.record.response_string import ResponseString

RAW_FORMAT = 'raw'

def print_raw_response(response: requests.Response, file_path = None):
  mitmproxy_response = PythonResponseAdapterFactory(response).mitmproxy_response()
  facade = MitmproxyResponseFacade(mitmproxy_response)
  response_string = ResponseString(facade, None)

  if not file_path:
    print(response_string.get().decode(), end="")
  else:
    with open(file_path, 'w') as fp:
      fp.write(response_string.get().decode()) 


import pathlib
import pdb

from click.testing import CliRunner

from stoobly_agent.app.cli import scaffold
from stoobly_agent.config.data_dir import DATA_DIR_NAME


class ScaffoldCliInvoker():

  @staticmethod
  def cli_app_create(runner: CliRunner, app_dir_path: str, app_name: str):
    pathlib.Path(f"{app_dir_path}/{DATA_DIR_NAME}").mkdir(parents=True, exist_ok=True)

    result = runner.invoke(scaffold, ['app', 'create',
      '--app-dir-path', app_dir_path,
      '--force',
      app_name
    ])

    assert result.exit_code == 0
    output = result.stdout
    assert not output

  @staticmethod
  def cli_app_mkcert(runner: CliRunner, app_dir_path: str):
    result = runner.invoke(scaffold, ['app', 'mkcert',
      '--app-dir-path', app_dir_path,
      '--context-dir-path', app_dir_path,
    ])

    assert result.exit_code == 0
    output = result.stdout
    assert not output

  @staticmethod
  def cli_service_create(runner: CliRunner, app_dir_path: str, hostname: str, service_name: str, https: bool):
    scheme = 'http'
    port = '80'
    if https == True:
      scheme = 'https'
      port = '443'

    result = runner.invoke(scaffold, ['service', 'create',
      '--app-dir-path', app_dir_path,
      '--env', 'TEST',
      '--force',
      '--hostname', hostname,
      '--scheme', scheme,
      '--port', port,
      '--workflow', 'mock',
      '--workflow', 'record',
      '--workflow', 'test',
      service_name
    ])
    assert result.exit_code == 0
    output = result.stdout
    assert not output

  # Specific flags for assets
  @staticmethod
  def cli_service_create_assets(runner: CliRunner, app_dir_path: str, hostname: str, service_name: str, https: bool):
    scheme = 'http'
    port = '80'
    if https == True:
      scheme = 'https'
      port = '443'
    proxy_mode_reverse_spec = f"reverse:{scheme}://{hostname}:8080"

    result = runner.invoke(scaffold, ['service', 'create',
      '--app-dir-path', app_dir_path,
      '--force',
      '--hostname', hostname,
      '--scheme', scheme,
      '--port', port,
      '--proxy-mode', proxy_mode_reverse_spec,
      '--detached',
      '--workflow', 'test',
      service_name
    ])
    assert result.exit_code == 0
    output = result.stdout
    assert not output

  @staticmethod
  def cli_service_delete(runner: CliRunner, app_dir_path: str, service_name: str):
    result = runner.invoke(scaffold, ['service', 'delete',
      '--app-dir-path', app_dir_path,
      service_name
    ])
    assert result.exit_code == 0
    output = result.stdout
    assert 'error' not in output.lower()

  @staticmethod
  def cli_workflow_create(runner: CliRunner, app_dir_path: str, service_name: str):
    result = runner.invoke(scaffold, ['workflow', 'create',
      '--app-dir-path', app_dir_path,
      '--service', service_name,
      '--template', 'mock',
      'ci',
    ])

    assert result.exit_code == 0
    output = result.stdout
    assert not output

  @staticmethod
  def cli_workflow_up(runner: CliRunner, app_dir_path: str, target_workflow_name: str):
    command = ['workflow', 'up',
      '--app-dir-path', app_dir_path,
      '--context-dir-path', app_dir_path,
      target_workflow_name,
    ]
    result = runner.invoke(scaffold, command)

    assert result.exit_code == 0
    output = result.stdout
    assert output

  @staticmethod
  def cli_workflow_down(runner: CliRunner, app_dir_path: str, target_workflow_name: str):
    command = ['workflow', 'down',
      '--app-dir-path', app_dir_path,
      '--context-dir-path', app_dir_path,
      target_workflow_name,
    ]
    result = runner.invoke(scaffold, command)

    assert result.exit_code == 0
    output = result.stdout
    assert output


import click
import os
import pdb
import re
import sys

from io import TextIOWrapper
from typing import List

from stoobly_agent.app.cli.helpers.certificate_authority import CertificateAuthority
from stoobly_agent.app.cli.helpers.shell import exec_stream
from stoobly_agent.app.cli.scaffold.app import App
from stoobly_agent.app.cli.scaffold.app_create_command import AppCreateCommand
from stoobly_agent.app.cli.scaffold.constants import (
  DOCKER_NAMESPACE, WORKFLOW_CONTAINER_PROXY, WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE
)
from stoobly_agent.app.cli.scaffold.containerized_app import ContainerizedApp
from stoobly_agent.app.cli.scaffold.docker.service.configure_gateway import configure_gateway
from stoobly_agent.app.cli.scaffold.docker.workflow.decorators_factory import get_workflow_decorators
from stoobly_agent.app.cli.scaffold.hosts_file_manager import HostsFileManager
from stoobly_agent.app.cli.scaffold.service import Service
from stoobly_agent.app.cli.scaffold.service_config import ServiceConfig
from stoobly_agent.app.cli.scaffold.service_create_command import ServiceCreateCommand
from stoobly_agent.app.cli.scaffold.service_delete_command import ServiceDeleteCommand
from stoobly_agent.app.cli.scaffold.service_workflow_validate_command import ServiceWorkflowValidateCommand
from stoobly_agent.app.cli.scaffold.templates.constants import CORE_SERVICES
from stoobly_agent.app.cli.scaffold.validate_exceptions import ScaffoldValidateException
from stoobly_agent.app.cli.scaffold.workflow import Workflow
from stoobly_agent.app.cli.scaffold.workflow_create_command import WorkflowCreateCommand
from stoobly_agent.app.cli.scaffold.workflow_copy_command import WorkflowCopyCommand
from stoobly_agent.app.cli.scaffold.workflow_log_command import WorkflowLogCommand
from stoobly_agent.app.cli.scaffold.workflow_run_command import WorkflowRunCommand
from stoobly_agent.app.cli.scaffold.workflow_validate_command import WorkflowValidateCommand
from stoobly_agent.config.constants import env_vars
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.logger import bcolors, DEBUG, ERROR, INFO, Logger, WARNING

from .helpers.print_service import FORMATS, print_services, select_print_options

LOG_ID = 'Scaffold'

current_working_dir = os.getcwd()
data_dir: DataDir = DataDir.instance()

@click.group(
    epilog="Run 'stoobly-agent project COMMAND --help' for more information on a command.",
    help="Manage scaffold"
)
@click.pass_context
def scaffold(ctx):
    pass

@click.group(
  epilog="Run 'stoobly-agent request response COMMAND --help' for more information on a command.",
  help="Manage app scaffold"
)
@click.pass_context
def app(ctx):
    pass

@click.group(
  epilog="Run 'stoobly-agent request response COMMAND --help' for more information on a command.",
  help="Manage service scaffold"
)
@click.pass_context
def service(ctx):
    pass

@click.group(
  epilog="Run 'stoobly-agent request response COMMAND --help' for more information on a command.",
  help="Manage workflow scaffold"
)
@click.pass_context
def workflow(ctx):
    pass

@click.group(
  epilog="Run 'stoobly-agent scaffold hostname COMMAND --help' for more information on a command.",
  help="Manage scaffold service hostnames"
)
@click.pass_context
def hostname(ctx):
    pass

@app.command(
  help="Scaffold application"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to create the app scaffold.')
@click.option('--force', is_flag=True, help='Overwrite maintained scaffolded app files.')
@click.option('--network', help='App default network name. Defaults to app name.')
@click.argument('app_name')
def create(**kwargs):
  __validate_app_dir(kwargs['app_dir_path'])

  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)

  if kwargs['force'] or not os.path.exists(app.scaffold_namespace_path):
    if not kwargs['network']:
      kwargs['network'] = kwargs['app_name']

    AppCreateCommand(app, **kwargs).build()
  else:
    print(f"{kwargs['app_dir_path']} already exists, use option '--force' to continue ")

@app.command(
  help="Scaffold app service certs"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--ca-certs-dir-path', default=data_dir.ca_certs_dir_path, help='Path to ca certs directory used to sign SSL certs.')
@click.option('--certs-dir-path', help='Path to certs directory. Defaults to the certs dir of the context.')
@click.option('--context-dir-path', default=data_dir.context_dir_path, help='Path to Stoobly data directory.')
@click.option('--service', multiple=True, help='Select which services to run. Defaults to all.')
@click.option('--workflow', multiple=True, help='Specify services by workflow(s). Defaults to all.')
def mkcert(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE, **kwargs)
  __validate_app(app)

  services = __get_services(
    app, service=kwargs['service'], without_core=True, workflow=kwargs['workflow']
  )

  __services_mkcert(app, services)

@service.command(
  help="Scaffold a service",
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--detached', is_flag=True, help='Use isolated and non-persistent context directory.')
@click.option('--env', multiple=True, help='Specify an environment variable.')
@click.option('--force', is_flag=True, help='Overwrite maintained scaffolded service files.')
@click.option('--hostname', help='Service hostname.')
@click.option('--port', type=click.IntRange(1, 65535), help='Service port.')
@click.option('--priority', default=5, type=click.FloatRange(1.0, 9.0), help='Determines the service run order. Lower values run first.')
@click.option('--proxy-mode', help='''
  Proxy mode can be "regular", "transparent", "socks5",
  "reverse:SPEC", or "upstream:SPEC". For reverse and
  upstream proxy modes, SPEC is host specification in
  the form of "http[s]://host[:port]".
''')
@click.option('--scheme', type=click.Choice(['http', 'https']), help='Defaults to https if hostname is set.')
@click.option('--workflow', multiple=True, type=click.Choice([WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE]), help='Include pre-defined workflows.')
@click.argument('service_name')
def create(**kwargs):
  __validate_app_dir(kwargs['app_dir_path'])

  if '/' in kwargs['service_name']:
    print(f"Error: {kwargs['service_name']} is invalid. It cannot container '/", file=sys.stderr)
    sys.exit(1)

  if kwargs.get('hostname'):
    hostname_regex = re.compile(r'^[a-zA-Z0-9.-]+$')
    if not re.search(hostname_regex, kwargs['hostname']):
      print(f"Error: {kwargs['hostname']} is invalid.", file=sys.stderr)
      sys.exit(1)

  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)

  service = Service(kwargs['service_name'], app)
  if kwargs['force'] or not os.path.exists(service.dir_path):
    __scaffold_build(app, **kwargs)
  else:
    print(f"{service.dir_path} already exists, use option '--force' to continue")

@service.command(
  help="List services",
  name="list"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--format', type=click.Choice(FORMATS), help='Format output.')
@click.option('--select', multiple=True, help='Select column(s) to display.')
@click.option('--service', multiple=True, help='Select specific services.')
@click.option('--without-headers', is_flag=True, default=False, help='Disable printing column headers.')
@click.option('--workflow', multiple=True, help='Specify workflow(s) to filter the services by. Defaults to all.')
def _list(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  services = __get_services(app, service=kwargs['service'], workflow=kwargs['workflow'])

  rows = []
  for service_name in services: 
    service = Service(service_name, app)
    __validate_service_dir(service.dir_path)

    service_config = ServiceConfig(service.dir_path)
    rows.append({
      'name': service_name,
      **service_config.to_dict()
    })

  print_services(rows, **select_print_options(kwargs))

@service.command(
  help="Delete a service",
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.argument('service_name')
def delete(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  service = Service(kwargs['service_name'], app)

  if not os.path.exists(service.dir_path):
    print(f"Service does not exist, so not deleting")
  else:
    print(f"Deleting service: {service.service_name}")
    __scaffold_delete(app, **kwargs)
    print(f"Successfully deleted service: {service.service_name}")

@service.command(
  help="Update a service config"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--priority', help='Determines the service run order.')
@click.argument('service_name')
def update(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  service = Service(kwargs['service_name'], app)

  __validate_service_dir(service.dir_path)

  service_config = ServiceConfig(service.dir_path)

  if kwargs['priority']:
    service_config.priority = kwargs['priority']

  service_config.write()

@workflow.command(
  help="Create workflow for service(s)"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--context-dir-path', default=data_dir.context_dir_path, help='Path to Stoobly data directory.')
@click.option('--force', is_flag=True, help='Overwrite maintained scaffolded workflow files.')
@click.option('--service', multiple=True, help='Specify the service(s) to create the workflow for.')
@click.option('--template', required=True, type=click.Choice([WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE]), help='Select which workflow to use as a template.')
@click.argument('workflow_name')
def create(**kwargs):
  __validate_app_dir(kwargs['app_dir_path'])

  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE, **kwargs)

  for service_name in kwargs['service']:
    config = { **kwargs }
    del config['service']
    config['service_name'] = service_name

    service = Service(service_name, app)
    __validate_service_dir(service.dir_path)

    workflow_dir_path = service.workflow_dir_path(kwargs['workflow_name'])
    if kwargs['force'] or not os.path.exists(workflow_dir_path):
      __workflow_create(app, **config)
    else:
      print(f"{workflow_dir_path} already exists, use option '--force' to continue")

@workflow.command(
  help="Copy a workflow for service(s)",
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--context-dir-path', default=data_dir.context_dir_path, help='Path to Stoobly data directory.')
@click.option('--service', multiple=True, help='Specify service(s) to add the workflow to.')
@click.argument('workflow_name')
@click.argument('destination_workflow_name')
def copy(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE, **kwargs)

  for service_name in kwargs['service']:
    config = { **kwargs }
    del config['service']
    config['service_name'] = service_name
      
    command = WorkflowCopyCommand(app, **config)

    if not command.app_dir_exists:
      print(f"Error: {command.app_dir_path} does not exist", file=sys.stderr)
      sys.exit(1)

    command.copy(kwargs['destination_workflow_name'])

@workflow.command()
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--context-dir-path', default=data_dir.context_dir_path, help='Path to Stoobly data directory.')
@click.option('--containerized', is_flag=True, help='Set if run from within a container.')
@click.option('--dry-run', default=False, is_flag=True)
@click.option('--extra-entrypoint-compose-path', help='Path to extra entrypoint compose file.')
@click.option('--log-level', default=INFO, type=click.Choice([DEBUG, INFO, WARNING, ERROR]), help='''
    Log levels can be "debug", "info", "warning", or "error"
''')
@click.option('--namespace', help='Workflow namespace.')
@click.option('--network', help='Workflow network name.')
@click.option('--rmi', is_flag=True, help='Remove images used by containers.')
@click.option('--script-path', help='Path to intermediate script path.')
@click.option('--service', multiple=True, help='Select which services to log. Defaults to all.')
@click.option('--user-id', default=os.getuid(), help='OS user ID of the owner of context dir path.')
@click.argument('workflow_name')
def down(**kwargs):  
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE, **kwargs)
  __validate_app(app)

  # If namespace is set, default network to namespace
  if kwargs['namespace'] and not kwargs['network']:
    kwargs['network'] = kwargs['namespace']

  services = __get_services(
    app, service=kwargs['service'], workflow=[kwargs['workflow_name']]
  )

  commands: List[WorkflowRunCommand] = []
  for service in services:
    config = { **kwargs }
    config['service_name'] = service
    command = WorkflowRunCommand(app, **config)
    command.current_working_dir = current_working_dir
    commands.append(command)

  script = __build_script(**kwargs)

  commands = sorted(commands, key=lambda command: command.service_config.priority)
  for index, command in enumerate(commands):
    __print_header(f"SERVICE {command.service_name}")

    extra_compose_path = None

    # By default, the entrypoint service should be last
    # However, this can change if the user has configured a service's priority to be higher
    if index == len(commands) - 1:
      extra_compose_path = kwargs['extra_entrypoint_compose_path']

    exec_command = command.down(
      extra_compose_path=extra_compose_path,
      namespace=kwargs['namespace'],
      rmi=kwargs['rmi'],
      user_id=kwargs['user_id']
    )
    if not exec_command:
      continue
    
    print(exec_command, file=script)

  # After services are stopped, their network needs to be removed
  if len(commands) > 0:
    command: WorkflowRunCommand = commands[0]

    if kwargs['rmi']:
      remove_image_command = command.remove_image(kwargs['user_id'])
      print(remove_image_command, file=script)

    remove_egress_network_command = command.remove_egress_network()
    print(remove_egress_network_command, file=script)

    remove_ingress_network_command = command.remove_ingress_network()
    print(remove_ingress_network_command, file=script)

  __run_script(script, kwargs['dry_run'])

  # Options are no longer valid
  if kwargs['containerized'] and os.path.exists(data_dir.mitmproxy_options_json_path):
    os.remove(data_dir.mitmproxy_options_json_path)

@workflow.command()
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option(
  '--container', multiple=True, help=f"Select which containers to log. Defaults to '{WORKFLOW_CONTAINER_PROXY}'"
)
@click.option('--dry-run', default=False, is_flag=True, help='If set, prints commands.')
@click.option('--follow', is_flag=True, help='Follow last container log output.')
@click.option('--log-level', default=INFO, type=click.Choice([DEBUG, INFO, WARNING, ERROR]), help='''
    Log levels can be "debug", "info", "warning", or "error"
''')
@click.option('--namespace', help='Workflow namespace.')
@click.option('--script-path', help='Path to intermediate script path.')
@click.option('--service', multiple=True, help='Select which services to log. Defaults to all.')
@click.argument('workflow_name')
def logs(**kwargs):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  if len(kwargs['container']) == 0:
    kwargs['container'] = [WORKFLOW_CONTAINER_PROXY]

  services = __get_services(
    app, service=kwargs['service'], without_core=True, workflow=[kwargs['workflow_name']]
  )

  commands: List[WorkflowLogCommand] = []
  for service in services:
    if len(kwargs['service']) == 0:
      # If no filter is specified, ignore CORE_SERVICES
      if service in CORE_SERVICES:
        continue
    else:
      # If a filter is specified, ignore all other services
      if service not in kwargs['service']:
        continue

    config = { **kwargs }
    config['service_name'] = service 
    command = WorkflowLogCommand(app, **config)
    commands.append(command)

  script = __build_script(**kwargs)

  commands = sorted(commands, key=lambda command: command.service_config.priority)
  for index, command in enumerate(commands):
    __print_header(f"SERVICE {command.service_name}")

    follow = kwargs['follow'] and index == len(commands) - 1
    shell_commands = command.build(
      containers=kwargs['container'], follow=follow, namespace=kwargs['namespace']
    )

    for shell_command in shell_commands:
      print(shell_command, file=script)

  __run_script(script, kwargs['dry_run'])

@workflow.command()
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--ca-certs-dir-path', default=data_dir.ca_certs_dir_path, help='Path to ca certs directory used to sign SSL certs.')
@click.option('--certs-dir-path', help='Path to certs directory. Defaults to the certs dir of the context.')
@click.option('--containerized', is_flag=True, help='Set if run from within a container.')
@click.option('--context-dir-path', default=data_dir.context_dir_path, help='Path to Stoobly data directory.')
@click.option('--detached', is_flag=True, help='If set, will not run the highest priority service in the foreground.')
@click.option('--dry-run', default=False, is_flag=True, help='If set, prints commands.')
@click.option('--extra-entrypoint-compose-path', help='Path to extra entrypoint compose file.')
@click.option('--log-level', default=INFO, type=click.Choice([DEBUG, INFO, WARNING, ERROR]), help='''
    Log levels can be "debug", "info", "warning", or "error"
''')
@click.option('--mkcert', is_flag=True, help='Set to generate SSL certs for HTTPS services.')
@click.option('--namespace', help='Workflow namespace.')
@click.option('--network', help='Workflow network name.')
@click.option('--no-build', is_flag=True, help='Do not build images before starting containers.')
@click.option('--no-publish', is_flag=True, help='Do not publish all ports.')
@click.option('--pull', is_flag=True, help='Pull image before running.')
@click.option('--script-path', help='Path to intermediate script path.')
@click.option('--service', multiple=True, help='Select which services to run. Defaults to all.')
@click.option('--user-id', default=os.getuid(), help='OS user ID of the owner of context dir path.')
@click.option('--verbose', is_flag=True)
@click.option('--without-base', is_flag=True, help='Disable building Stoobly base image.')
@click.argument('workflow_name')
def up(**kwargs):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  containerized = kwargs['containerized']

  # Because we are running a docker-compose command which depends on APP_DIR env var
  # when we are running this command within a container, the host's app_dir_path will likely differ
  app_dir_path = current_working_dir if containerized else kwargs['app_dir_path']
  app = App(app_dir_path, DOCKER_NAMESPACE, **kwargs)
  __validate_app(app)

  # If namespace is set, default network to namespace
  if kwargs['namespace'] and not kwargs['network']:
    kwargs['network'] = kwargs['namespace']

  services = __get_services(
    app, service=kwargs['service'], workflow=[kwargs['workflow_name']]
  )

  if kwargs['mkcert']:
    _app = ContainerizedApp(app_dir_path, DOCKER_NAMESPACE) if containerized else app
    __services_mkcert(_app, services)

  # Gateway ports are dynamically set depending on the workflow run
  workflow = Workflow(kwargs['workflow_name'], app)
  configure_gateway(workflow.service_paths_from_services(services), kwargs['no_publish'])

  commands: List[WorkflowRunCommand] = []
  for service in services:
    config = { **kwargs }
    config['service_name'] = service
    command = WorkflowRunCommand(app, **config)
    command.current_working_dir = current_working_dir
    commands.append(command)

  script = __build_script(**kwargs)

  # Before services can be started, their image and network needs to be created
  if len(commands) > 0:
    command: WorkflowRunCommand = commands[0]

    init_commands = []
    if not kwargs['without_base']:
      create_image_command = command.create_image(user_id=kwargs['user_id'], verbose=kwargs['verbose'])
      init_commands.append(create_image_command)

    init_commands.append(command.create_egress_network())
    init_commands.append(command.create_ingress_network())
    joined_command = ' && '.join(init_commands)

    if not containerized:
      command.write_nameservers()

    print(joined_command, file=script)

  commands = sorted(commands, key=lambda command: command.service_config.priority)
  for index, command in enumerate(commands):
    __print_header(f"SERVICE {command.service_name}")

    attached = False
    extra_compose_path = None

    # By default, the entrypoint service should be last
    # However, this can change if the user has configured a service's priority to be higher
    if index == len(commands) - 1:
      attached = not kwargs['detached']
      extra_compose_path = kwargs['extra_entrypoint_compose_path']

    exec_command = command.up(
      attached=attached,
      extra_compose_path=extra_compose_path,
      namespace=kwargs['namespace'],
      no_build=kwargs['no_build'],
      pull=kwargs['pull'],
      user_id=kwargs['user_id']
    )
    if not exec_command:
      continue

    print(exec_command, file=script)

  __run_script(script, kwargs['dry_run'])


@workflow.command(
  help="Validate a scaffold workflow"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to validate the app scaffold.')
@click.argument('workflow_name')
def validate(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  workflow = Workflow(kwargs['workflow_name'], app)
  
  config = { **kwargs }
  config['service_name'] = 'build'

  try:
    command = WorkflowValidateCommand(app, **config)
    command.validate()
  except ScaffoldValidateException as sve:
    print(f"{bcolors.FAIL}\nFatal scaffold validation exception:{bcolors.ENDC}\n{sve}", file=sys.stderr)
    print("\nSee the scaffold workflow troubleshooting guide at: https://docs.stoobly.com/guides/how-to-integrate-e2e-testing/how-to-run-a-workflow/troubleshooting", file=sys.stderr)
    sys.exit(1)

  try:
    for service in workflow.services_ran:
      if service not in CORE_SERVICES:
        config['service_name'] = service
        command = ServiceWorkflowValidateCommand(app, **config)
        command.validate()
  except ScaffoldValidateException as sve:
    print(f"{bcolors.FAIL}\nFatal scaffold validation exception:{bcolors.ENDC}\n{sve}", file=sys.stderr)
    print("\nSee the scaffold workflow troubleshooting guide at: https://docs.stoobly.com/guides/how-to-integrate-e2e-testing/how-to-run-a-workflow/troubleshooting", file=sys.stderr)
    sys.exit(1)

  print(f"{bcolors.OKCYAN}✔ Done validating Stoobly scaffold and services, success!{bcolors.ENDC}")

@hostname.command(
  help="Update the system hosts file for all scaffold service hostnames"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--service', multiple=True, help='Select specific services. Defaults to all.')
@click.option('--workflow', multiple=True, help='Specify services by workflow(s). Defaults to all.')
def install(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  services = __get_services(
    app, service=kwargs['service'], without_core=True, workflow=kwargs['workflow']
  )

  hostnames = []
  for service_name in services: 
    service = Service(service_name, app)
    __validate_service_dir(service.dir_path)

    service_config = ServiceConfig(service.dir_path)
    if service_config.hostname:
      hostnames.append(service_config.hostname)

  __elevate_sudo()

  try:
    hosts_file_manager = HostsFileManager()
    hosts_file_manager.install_hostnames(hostnames)
  except PermissionError:
    print("Permission denied. Please run this command with sudo.", file=sys.stderr)
    sys.exit(1)

@hostname.command(
  help="Delete from the system hosts file all scaffold service hostnames"
)
@click.option('--app-dir-path', default=current_working_dir, help='Path to application directory.')
@click.option('--service', multiple=True, help='Select specific services. Defaults to all.')
@click.option('--workflow', multiple=True, help='Specify services by workflow(s). Defaults to all.')
def uninstall(**kwargs):
  app = App(kwargs['app_dir_path'], DOCKER_NAMESPACE)
  __validate_app(app)

  services = __get_services(
    app, service=kwargs['service'], without_core=True, workflow=kwargs['workflow']
  )

  hostnames = []
  for service_name in services: 
    service = Service(service_name, app)
    __validate_service_dir(service.dir_path)

    service_config = ServiceConfig(service.dir_path)
    if service_config.hostname:
      hostnames.append(service_config.hostname)

  __elevate_sudo()

  try:
    hosts_file_manager = HostsFileManager()
    hosts_file_manager.uninstall_hostnames(hostnames)
  except PermissionError:
    print("Permission denied. Please run this command with sudo.", file=sys.stderr)
    sys.exit(1)

scaffold.add_command(app)
scaffold.add_command(service)
scaffold.add_command(workflow)
scaffold.add_command(hostname)

def __build_script(**kwargs):
  script_path = kwargs['script_path']
  if not script_path:
    script_file_name = f"{kwargs['workflow_name']}.sh"
    script_path = os.path.join(data_dir.tmp_dir_path, script_file_name)
  else:
    script_dir = os.path.dirname(script_path)
    if not os.path.exists(script_dir):
      os.makedirs(script_dir, exist_ok=True)

  # Truncate
  with open(script_path, 'w'):
    pass

  return open(script_path, 'a')

def __elevate_sudo():
  import subprocess

  if os.geteuid() != 0:
    subprocess.run(["sudo", sys.executable] + sys.argv)
    sys.exit(0)

def __get_services(app: App, **kwargs):
  selected_services = list(kwargs['service'])

  if not selected_services:
    selected_services = app.services
  else:
    selected_services += CORE_SERVICES
    missing_services = [service for service in selected_services if service not in app.services]

    if missing_services:
      # Warn if an invalid service is provided
      Logger.instance(LOG_ID).warn(f"Service(s) {','.join(missing_services)} are not found")

    # Remove services that don't exist
      selected_services = list(set(selected_services) - set(missing_services))

  # If without_score is set, filter out CORE_SERVICES
  if kwargs.get('without_core'):
    selected_services = list(set(selected_services) - set(CORE_SERVICES))

  # If workflow is set, keep only services in the workflow
  if kwargs.get('workflow'):
    workflow_services = []
    for workflow_name in kwargs['workflow']:
      workflow = Workflow(workflow_name, app)
      workflow_services += workflow.services

    # Intersection
    selected_services = list(filter(lambda x: x in workflow_services, selected_services))

  services = list(set(selected_services))
  services.sort()
  
  return services

def __print_header(text: str):
  Logger.instance(LOG_ID).info(f"{bcolors.OKBLUE}{text}{bcolors.ENDC}")

def __run_script(script: TextIOWrapper, dry_run = False):
  script.close()

  with open(script.name, 'r') as fp:
    for line in fp:
      if not dry_run:
        exec_stream(line.strip())
      else:
        print(line.strip())

def __scaffold_build(app, **kwargs):
  command = ServiceCreateCommand(app, **kwargs)

  command.build()

def __scaffold_delete(app, **kwargs):
  command = ServiceDeleteCommand(app, **kwargs)

  command.delete()

def __services_mkcert(app: App, services):
  for service_name in services:
    service = Service(service_name, app)
    __validate_service_dir(service.dir_path)

    service_config = ServiceConfig(service.dir_path)

    if service_config.scheme != 'https':
      continue

    hostname = service_config.hostname
    
    if not hostname:
      continue

    ca = CertificateAuthority(app.ca_certs_dir_path)
    if not ca.signed(hostname, app.certs_dir_path):
      Logger.instance(LOG_ID).info(f"Creating cert for {hostname}")
      ca.sign(hostname, app.certs_dir_path)

def __validate_app(app: App):
  if not app.valid:
    print(f"Error: {app.dir_path} is not a valid scaffold app", file=sys.stderr)
    sys.exit(1)

def __validate_app_dir(app_dir_path):
  if not os.path.exists(app_dir_path):
    print(f"Error: {app_dir_path} does not exist", file=sys.stderr)
    sys.exit(1)

def __validate_service_dir(service_dir_path):
  if not os.path.exists(service_dir_path):
    print(f"Error: {service_dir_path} does not exist, please scaffold this service", file=sys.stderr)
    sys.exit(1)

def __workflow_create(app, **kwargs):
  command = WorkflowCreateCommand(app, **kwargs)

  service_config = command.service_config
  workflow_decorators = get_workflow_decorators(kwargs['template'], service_config)

  command.build(
    template=kwargs['template'],
    workflow_decorators=workflow_decorators
  )


import pdb
import threading

from typing import TypedDict

from stoobly_agent.config.constants.statuses import REQUESTS_MODIFIED
from stoobly_agent.lib.api.agent_api import AgentApi
from stoobly_agent.lib.cache import Cache
from stoobly_agent.lib.logger import Logger

class Options(TypedDict):
  sync: bool

# Announce that a new request has been created
def publish_change(status: str, value: any, **options: Options):
  if options.get('sync'):
    return __publish_change_sync(status, value)

  from stoobly_agent.app.settings import Settings
  settings = Settings.instance()

  # If ui is not active, return
  if not settings.ui.active:
    return False

  ui_url = settings.ui.url

  if not ui_url:
    Logger.instance().warn('Settings.ui.url not configured')
    return False
  else:
    thread = threading.Thread(target=__put_status, args=(ui_url, status, value))
    thread.start()
    return True

def publish_requests_modified(value, **options: Options):
  return publish_change(REQUESTS_MODIFIED, value, **options)

def __publish_change_sync(status: str, value):
  cache = Cache.instance()
  cache.write(status, value)
  return True

def __put_status(ui_url, status, value):
    api: AgentApi = AgentApi(ui_url)
    api.update_status(status, value)


import os
import pdb
import requests

from stoobly_agent.config.constants.env_vars import HTTP_PROXY, HTTPS_PROXY, NO_PROXY
from stoobly_agent.app.settings import Settings

class Api():

    def without_proxy(self, handler):
      settings = Settings.instance()

      no_proxy = os.environ.get(NO_PROXY)
      os.environ[NO_PROXY] = settings.remote.api_url

      current_proxies = self.set_proxy('')

      res = handler()

      for key, val in current_proxies.items():
        if not val:
          continue
        
        if isinstance(val, str):
          os.environ[key] = val
        
      if isinstance(no_proxy, str): 
        os.environ[NO_PROXY] = no_proxy

      return res

    def with_proxy(self, handler):
      settings = Settings.instance()
      proxy_url = settings.proxy.url
      current = self.set_proxy(proxy_url)

      res = handler()

      for key, val in current.items():
        os.environ[key] = val

      return res

    def set_proxy(self, val):
      current = {}

      current[HTTP_PROXY] = os.environ.get(HTTP_PROXY) or ''
      os.environ[HTTP_PROXY] = val

      current[HTTPS_PROXY] = os.environ.get(HTTPS_PROXY) or ''
      os.environ[HTTPS_PROXY] = val

      current[HTTP_PROXY.lower()] = os.environ.get(HTTP_PROXY.lower()) or ''
      os.environ[HTTP_PROXY.lower()] = val

      current[HTTPS_PROXY.lower()] = os.environ.get(HTTPS_PROXY.lower()) or ''
      os.environ[HTTPS_PROXY.lower()] = val

      return current

    def get(self, url, **kwargs):
      handler = lambda: requests.get(url, **kwargs)
      return self.without_proxy(handler)

    def post(self, url, **kwargs):
      handler = lambda: requests.post(url, **kwargs)
      return self.without_proxy(handler)

    def put(self, url, **kwargs):
      handler = lambda: requests.put(url, **kwargs)
      return self.without_proxy(handler)
    
    def delete(self, url, **kwargs):
      handler = lambda: requests.delete(url, **kwargs)
      return self.without_proxy(handler)



import pdb

from typing import TypedDict

from ...constants import DOCKER_NAMESPACE
from ..constants import DOCKERFILE_SERVICE
from .builder import ServiceBuilder
from .types import BuildDecoratorOptions

class BuildDecorator():

  def __init__(self, service_builder: ServiceBuilder):
    self.__service_builder = service_builder

  def decorate(self, **kwargs: BuildDecoratorOptions):
    service_builder = self.__service_builder
    build = {
      'context': '../..', # Assumes app root is 2 levels up
      'dockerfile': f"./{DOCKER_NAMESPACE}/{service_builder.service_name}/{DOCKERFILE_SERVICE}"
    }

    if 'build_args' in kwargs:
      args = {}
      for arg in kwargs['build_args']:
        args[arg] = '${' + arg + '}'
      build['args'] = args

    services = self.__service_builder.services
    app_name = self.__service_builder.app_base
    app_service = services.get(app_name) or {}

    services[app_name] = { 
      **app_service,
      **{ 
          'build': build,
        } 
    }


import pdb
import requests

from io import BytesIO
from mitmproxy.http import Request as MitmproxyRequest
from urllib3 import HTTPResponse
from urllib3.exceptions import InsecureRequestWarning

from stoobly_agent.app.proxy.mitmproxy.flow_mock import MitmproxyFlowMock
from stoobly_agent.app.models.adapters.mitmproxy import MitmproxyRequestAdapterFactory, MitmproxyResponseAdapterFactory

from .intercept_handler import request as handle_request, response as handle_response

def simulate_intercept(request: requests.Request, **config):
  flow = MitmproxyFlowMock()

  flow.request = request

  handle_request(flow)

  if not flow.response:
    session = requests.Session()
    prepared_request = __prepare_mitmproxy_request(flow.request)

    res = None 

    if not config.get('verify'):
      # Suppress only the single warning from urllib3 needed.
      requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

    try:
      res = session.send(prepared_request, **{'timeout': 300, **config})
    except requests.exceptions.ConnectTimeout:
      res = __response(b'Gateway Timeout', 504)
    except requests.exceptions.SSLError as e:
      res = __response(str(e).encode(), 502)
    except requests.exceptions.ConnectionError:
      res = __response(b'Bad Gateway', 502)
    except Exception:
      res = __response(b'Unknown Error', 0)

    flow.response = res

  handle_response(flow)

  return MitmproxyResponseAdapterFactory(flow.response).python_response()

def __prepare_mitmproxy_request(request: MitmproxyRequest) -> requests.Request:
  _request = MitmproxyRequestAdapterFactory(request).python_request()
  return _request.prepare()

def __response(body: bytes, status: int, headers = {}):
  res = requests.Response()

  res.status_code = status
  res.headers = headers
  res.raw = HTTPResponse(
    body=BytesIO(body),
    decode_content=False,
    headers=headers,
    preload_content=False
  )

  return res


import os
import pdb
import requests
import sys

from stoobly_agent.app.cli.helpers.handle_replay_service import DEFAULT_FORMAT, handle_before_replay, handle_after_replay, print_session, ReplaySession
from stoobly_agent.app.cli.helpers.handle_test_service import SessionContext, exit_on_failure, handle_test_complete, handle_test_session_complete
from stoobly_agent.app.cli.helpers.test_facade import TestFacade
from stoobly_agent.app.models.helpers.apply import Apply
from stoobly_agent.app.proxy.replay.body_parser_service import decode_response
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import alias_resolve_strategy, env_vars
from stoobly_agent.lib.api.keys.request_key import InvalidRequestKey
from stoobly_agent.lib.utils import jmespath

from ..helpers.print_service import print_requests, select_print_options
from ..helpers.request_facade import RequestFacade
from ..helpers.validations import *
from ..types.request import RequestTestOptions

def delete_handler(kwargs):
  validate_request_key(kwargs['request_key'])

  request = RequestFacade(Settings.instance())
  res, status = request.show(kwargs['request_key'], {})

  if filter_response(res, status):
    sys.exit(1)

  is_deleted = res.get('is_deleted')
  if not res:
    print(f"Error: Could not find request", file=sys.stderr)
    sys.exit(1)

  res = request.delete(kwargs['request_key'])
  if not res:
    print('Error: Could not delete request', file=sys.stderr)
    sys.exit(1)

  if not is_deleted:
    print('Request moved to trash!')
  else:
    print('Request deleted!')

def get_handler(kwargs):
  validate_request_key(kwargs['request_key'])

  request = RequestFacade(Settings.instance())
  res = __replay(request.mock, kwargs)

  print(res.text)

def list_handler(kwargs):
  print_options = select_print_options(kwargs)

  settings = Settings.instance()
  project_key = None

  if kwargs.get('is_remote'):
    project_key = resolve_project_key_and_validate(kwargs, settings)
    del kwargs['project_key']

  if kwargs.get('scenario_key'):
    validate_scenario_key(kwargs['scenario_key'])

  if kwargs.get('search'):
    kwargs['q'] = kwargs['search']
    del kwargs['search']

  request = RequestFacade(settings)
  requests_response, status = request.index(project_key, kwargs)

  if filter_response(requests_response, status):
    sys.exit(1)

  if not requests_response or len(requests_response['list']) == 0:
    print('No requests found.')
  else:
    print_requests(requests_response['list'], **print_options)

def query_handler(kwargs):
  validate_request_key(kwargs['request_key'])

  request = RequestFacade(Settings.instance())
  res = __replay(request.mock, kwargs)

  content = res.content
  content_type = res.headers.get('content-type')
  decoded_response = decode_response(content, content_type)

  print(jmespath.search(kwargs['query'], decoded_response))

def replay_handler(kwargs):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  validate_request_key(kwargs['request_key'])

  if kwargs.get('scenario_key'):
    if not kwargs.get('record'):
      print("Error: Missing option '--record'.", file=sys.stderr)
      sys.exit(1)

    validate_scenario_key(kwargs['scenario_key'])

  if 'validate' in kwargs and len(kwargs['validate']):
      validate_aliases(kwargs['validate'], assign=kwargs['assign'], format=kwargs['format'], trace_id=kwargs['trace_id'])

  __assign_default_alias_resolve_strategy(kwargs)
  __assign_default_origin(kwargs)

  session: ReplaySession = {
    'buffer': kwargs['format'] and kwargs['format'] != DEFAULT_FORMAT,
    'contexts': [],
    'format': kwargs['format'],
    'total': 0,
  }
  
  kwargs['before_replay'] = lambda context: handle_before_replay(context, session)
  kwargs['after_replay'] = lambda context: handle_after_replay(context, session)
  
  request = RequestFacade(Settings.instance())
  __replay(request.replay, kwargs)

  print_session(session)

def reset_handler(kwargs):
  request_key = kwargs['request_key']
  _request_key = validate_request_key(request_key)

  apply_service = Apply(force=kwargs['force']).with_logger(print)
  resetted = apply_service.request(_request_key.id)

  if not resetted:
    print('Successfully reset the request!')
  else:
    print('Could not reset the request.')

def snapshot_handler(kwargs):
  request_key = kwargs['request_key']
  del kwargs['request_key']

  validate_request_key(request_key)

  request = RequestFacade(Settings.instance())
  request.snapshot(request_key, kwargs)

def test_handler(kwargs: RequestTestOptions):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  settings = Settings.instance()
  request_key = validate_request_key(kwargs['request_key'])

  if kwargs.get('remote_project_key'):
    validate_project_key(kwargs['remote_project_key'])

  if kwargs.get('report_key'):
    if not kwargs.get('save'):
      print("Error: Missing --save option", file=sys.stderr)
      sys.exit(1)

    validate_report_key(kwargs['report_key'])

  if len(kwargs['validate']):
    validate_aliases(kwargs['validate'], assign=kwargs['assign'], format=kwargs['format'], trace_id=kwargs['trace_id'])

  __assign_default_alias_resolve_strategy(kwargs)
  __assign_default_origin(kwargs)

  session: ReplaySession = {
    'buffer': kwargs['format'] and kwargs['format'] != DEFAULT_FORMAT,
    'contexts': [],
    'format': kwargs['format'],
    'total': 0,
  }

  session_context: SessionContext = { 
      'aggregate_failures': kwargs['aggregate_failures'], 
      'passed': 0, 
      'project_id': request_key.project_id, 
      'skipped': 0,
      'test_facade': TestFacade(settings), 
      'total': 0 
  }

  kwargs['before_replay'] = lambda context: handle_before_replay(
    context, session
  )

  kwargs['after_replay'] = lambda context: handle_test_complete(
    context, session_context, format=kwargs['format'], output_level=kwargs['output_level']
  )

  request = RequestFacade(settings)
  __replay(request.test, kwargs)

  handle_test_session_complete(session_context, format=kwargs['format'])

  exit_on_failure(session_context, format=kwargs['format'])

def __replay(handler, kwargs) -> requests.Response:
  request_key = kwargs['request_key']
  del kwargs['request_key']

  try:
    return handler(request_key, kwargs) 
  except InvalidRequestKey:
    print('Error: Invalid request key.', file=sys.stderr)
    sys.exit(1)

def __assign_default_alias_resolve_strategy(kwargs):
    # If we have assigned values to aliases, it's likely we want to also have them resolved
    if 'assign' in kwargs and len(kwargs['assign']) > 0 and kwargs['alias_resolve_strategy'] == alias_resolve_strategy.NONE:
        kwargs['alias_resolve_strategy'] = alias_resolve_strategy.FIFO

def __assign_default_origin(kwargs):
    if not kwargs.get('host'):
        kwargs['host'] = os.environ.get(env_vars.AGENT_REPLAY_HOST)

    if not kwargs.get('scheme'):
        kwargs['scheme'] = os.environ.get(env_vars.AGENT_REPLAY_SCHEME)


import pdb

def length_match_error(path_key, expected_value, actual_value):
    if not expected_value:
        return False, f"Key '{path_key}' length did not match: got {len(actual_value)}"
    else:
        return False, f"Key '{path_key}' length did not match: got {len(actual_value)}, expected {len(expected_value)}"

def param_name_exists_error(path_key):
    return False, f"Extra key: expected {path_key} to not exist"

def param_name_missing_error(path_key):
    return False, f"Missing key: expected {path_key} to exist"

def type_match_error(path_key, expected_value, actual_value):
    return False, f"Key '{path_key}' type did not match: got {actual_value}, expected {expected_value}"

def valid_type_error(path_key, value, valid_types):
    valid_types = list(map(lambda t: str(t), valid_types))
    return False, f"Key '{path_key}' type did not match: got {type(value)}, expected valid types {', '.join(valid_types)}"

def value_match_error(path_key, expected_value, actual_value):
    if len(path_key) == 0:
        return False, f"Match error: got {actual_value}, expected {expected_value}"
    else:
        return False, f"Key '{path_key}' did not match: got {actual_value}, expected {expected_value}"


from .ca_cert_cli import ca_cert
from .config_cli import config
from .dev_tools_cli import dev_tools
from .endpoint_cli import endpoint
from .feature_cli import feature
from .intercept_cli import intercept
from .main_group import MainGroup
from .project_cli import project
from .report_cli import report
from .request_cli import request
from .scaffold_cli import scaffold
from .scenario_cli import scenario
from .snapshot_cli import snapshot
from .trace_cli import trace


import click
import pdb
import sys

from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import mode, mock_policy, record_policy, replay_policy
from stoobly_agent.lib.api.keys.project_key import ProjectKey

from .helpers.handle_config_update_service import handle_intercept_active_update, handle_policy_update

settings = Settings.instance()

mode_options = [mode.MOCK, mode.RECORD, mode.REPLAY]

if settings.cli.features.remote:
    mode_options.append(mode.TEST)

active_mode = settings.proxy.intercept.mode

def __get_policy_options(active_mode):
    if active_mode == mode.MOCK:
        return [mock_policy.ALL, mock_policy.FOUND]
    elif active_mode == mode.RECORD:
        return [record_policy.ALL, record_policy.FOUND, record_policy.NOT_FOUND,record_policy.OVERWRITE]
    elif active_mode == mode.REPLAY:
        return [replay_policy.ALL]
    elif active_mode == mode.TEST:
        return [mock_policy.FOUND]

policy_options = __get_policy_options(active_mode)

@click.group(
    epilog="Run 'stoobly-agent intercept COMMAND --help' for more information on a command.",
    help="Manage request intercept"
)
@click.pass_context
def intercept(ctx):
    pass

@intercept.command(
    help="Enable intercept"
)
def enable(**kwargs):
    settings = Settings.instance()

    settings.proxy.intercept.active = True

    handle_intercept_active_update(settings)

    settings.commit()

    print("Intercept enabled!")

@intercept.command(
    help="Disable intercept"
)
def disable(**kwargs):
    settings = Settings.instance()

    settings.proxy.intercept.active = False

    handle_intercept_active_update(settings)

    settings.commit()

    print("Intercept disabled!")

@intercept.command(
    help="Configure intercept"
)
@click.option('--mode', type=click.Choice(mode_options))
@click.option('--policy', type=click.Choice(policy_options))
def configure(**kwargs):
    settings = Settings.instance()

    if not kwargs['mode'] and not kwargs['policy']:
        print("Error: Missing an option")
        sys.exit(1)

    if kwargs['mode']:
        if settings.proxy.intercept.mode != kwargs['mode']:
            if settings.proxy.intercept.active:
                settings.proxy.intercept.active = False
                handle_intercept_active_update(settings)

            settings.proxy.intercept.mode = kwargs['mode']

            print(f"Updating intercept mode to {kwargs['mode']}")

    _mode = kwargs['mode'] or settings.proxy.intercept.mode 

    if kwargs['policy']:
        active_mode = settings.proxy.intercept.mode
        valid_policies = __get_policy_options(active_mode)

        if not kwargs['policy'] in valid_policies:
            print("Error: Valid policies for {active_mode} are {valid_policies}", file=sys.stderr)
            sys.exit(1)

        project_key = ProjectKey(settings.proxy.intercept.project_key)
        data_rule = settings.proxy.data.data_rules(project_key.id)
        
        if active_mode == mode.MOCK:
            data_rule.mock_policy = kwargs['policy']
        elif active_mode == mode.RECORD:
            data_rule.record_policy = kwargs['policy']
        elif active_mode == mode.REPLAY:
            data_rule.replay_policy = kwargs['policy']
        elif active_mode == mode.TEST:
            data_rule.test_policy = kwargs['policy']

        handle_policy_update(settings)

        print(f"Updating {_mode} policy to {kwargs['policy']}")

    settings.commit()

@intercept.command(
    help="Show intercept"
)
def show(**kwargs):
    settings = Settings.instance()

    _mode = settings.proxy.intercept.mode
    project_key = ProjectKey(settings.proxy.intercept.project_key)
    data_rule = settings.proxy.data.data_rules(project_key.id)
    policy = None

    if active_mode == mode.MOCK:
        policy = data_rule.mock_policy
    elif active_mode == mode.RECORD:
        policy = data_rule.record_policy
    elif active_mode == mode.REPLAY:
        policy = data_rule.replay_policy
    elif active_mode == mode.TEST:
        policy = data_rule.test_policy

    if not _mode:
        print('No intercept mode set')
    else:
        print(f"{_mode.capitalize()} with policy {policy} {'enabled' if settings.proxy.intercept.active else 'disabled'}")



from . import ORM 

class Base(ORM.instance().base):

  @classmethod
  def where_for(cls, **kwargs):
    q = cls
    for key, val in kwargs.items():
      q = q.where(key, val)
    return q

  @classmethod
  def find_by(cls, **kwargs):
    return cls.where_for(**kwargs).first()

  @classmethod
  def last(cls):
    return cls.order_by('id', 'desc').first()



from .types.proxy_settings import MatchRule as IMatchRule

class MatchRule:

  def __init__(self, match_rule: IMatchRule):
    self.__match_rule = match_rule

    self.__components = self.__match_rule.get('components') or []
    self.__methods = self.__match_rule.get('methods') or []
    self.__modes = self.__match_rule.get('modes') or []
    self.__pattern = self.__match_rule.get('pattern')

  @property 
  def components(self):
    return self.__components

  @property
  def methods(self):
    return self.__methods

  @property
  def modes(self):
    return self.__modes

  @property
  def pattern(self):
    return self.__pattern

  def to_dict(self):
    return {
      'components': self.__components,
      'methods': self.__methods,
      'modes': self.__modes,
      'pattern': self.__pattern,
    }


import pdb
import pytest

from stoobly_agent.app.proxy.replay.body_parser_service import decode_response, encode_response

@pytest.fixture
def multipart_string():
  return b"--ce560532019a77d83195f9e9873e16a1\r\nContent-Disposition: form-data; name=\"author\"\r\n\r\nJohn Smith\r\n--ce560532019a77d83195f9e9873e16a1\r\nContent-Disposition: form-data; name=\"file\"; filename=\"example2.txt\"\r\nContent-Type: text/plain\r\nExpires: 0\r\n\r\nHello World\r\n--ce560532019a77d83195f9e9873e16a1--\r\n"

@pytest.fixture
def content_type():
  return "multipart/form-data; boundary=ce560532019a77d83195f9e9873e16a1"

class TestMultipart():

  def test_decodes_response(self, multipart_string: bytes, content_type: str):
    multidict = decode_response(multipart_string, content_type)

    assert multidict.get('author') == 'John Smith'
    assert multidict.get('file') == 'Hello World'

  def test_encodes_response(self, content_type: str):
    expected_params = { 'author': 'John Smith', 'file': 'Hello World'}
    multipart_string = encode_response(expected_params, content_type)

    multidict = decode_response(multipart_string, content_type)
    for key in expected_params:
      assert multidict.get(key) == expected_params[key]
